# -*- coding: utf-8 -*-
"""
Last modified on Nov 11, 2024

@author: Hermann Zeyen <hermann.zeyen@universite-paris-saclay.fr>
         Université Paris-Saclay, France
"""

import sys
import os
from copy import deepcopy
from datetime import datetime
import numpy as np
from PyQt5 import QtWidgets
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.collections import PatchCollection
from matplotlib.patches import Rectangle
from ..in_out.dialog import dialog
from .potential_prism import Prism_calc as PP
from . import mag_grav_utilities as utils
from ..plotting.new_window import newWindow


class inversion():
    """
    class controlling inversion of data sets
    """

    def __init__(self, data, x, y, z, topo_int=None, earth=None, data_type="m",
                 line_pos=None, direction=None):
        """
        Initialization of inversion.

        Parameters
        ----------
        data : list of one or two 1D or 2D numpy float arrays
            Data to be inverted [nT of mGal, depending on data_type].
            For 2.5D inversion, data are along a line in a 1D array.
            For 3D inversion, gridded data are used and passed as 2D array
            If two sensors have been used, each data set is one entry of the
            list. If only one sensor is used, the list has only one element.
        x, y, z : lists of one or two 1D or 2D numpy float arrays
            Coordinates of all data points having each the same structure as
            "data".
            If y[0] == None, 2D inversion is done and all y coordinates are
            set to 0. I.e., the coordinates along the line are X-coordinates.
        earth: class Earth_mag object, optional. Default=None
            Contains the properties of the Earth's magnetic field
        data_type : str, optional. Default: "m"
            Type of data to be inverted: "m"=magnetic data, "g"=gravity data.
            May be capitals or not.
        line_pos : float, optional. Defuult: None
            For 2.5D inversion, coordinate of the line to be inverted (X or Y).
            Only used for plotting.
        direction : str, optional. Default: None

        """
        self.data_shape = data[0].shape
        self.data1_shape = data[0].shape
        self.data = data[0].flatten()
        self.line_pos = line_pos
        self.direction = direction
        self.n_data = len(self.data)
        self.n_data1 = len(self.data)
        self.x = x[0].flatten()
        self.z = z[0].flatten()
        if min(self.x) > 100000.:
            self.dfac = 0.001
        else:
            self.dfac = 1.
        self.topo_int = topo_int
        if topo_int is None:
            self.topo_flag = False
        else:
            self.topo_flag = True
        self.dim = 2
        if y[0] is None:
            self.y = np.zeros_like(self.x)
        else:
            self.y = y[0].flatten()
            self.dim = 3
        # self.sensor1_data = self.n_data
# If only one data set is to be inverted, set values of second one to None
        if len(data) == 1:
            self.data2 = None
            self.x2 = None
            self.y2 = None
            self.z2 = None
            self.n_sensor = 1
            # self.sensor2_data = 0
            self.n_data2 = 0
# If two data sets are available, copy data and coordinates into their
# respective arrays
        else:
            self.data2_shape = data[1].shape
            self.data2 = data[1].flatten()
            self.x2 = x[1].flatten()
            self.z2 = z[1].flatten()
            if self.dim == 3:
                self.y2 = y[1].flatten()
            else:
                self.y2 = np.zeros_like(self.x2)
            self.n_sensor = 2
            # self.sensor2_data = len(self.data2)
            self.n_data2 = len(self.data2)
        self.data_type = data_type.lower()
        self.earth = earth
        self.sus_inv = False
        self.rem_inv = False
        self.rho_inv = False
        self.positive = False
        self.xprism_min = 0.
        self.xprism_max = 0.
        self.dx_prism = 0.
        self.min_size_x = 0.
        self.zprism_min = 0.
        self.zprism_max = 0.
        self.dz_prism = 0.
        self.min_size_z = 0.
        self.yprism_min = 0.
        self.yprism_max = 0.
        self.dy_prism = 0.
        self.min_size_y = 0.
        self.sigma_mag = 1.
        self.sigma_grav = 1.
        self.sigma_sus = 0.001
        self.sigma_rem = 0.04
        self.sigma_rho = 1.
        self.max_iter = 10
        self.max_diff_fac = 0.01
        self.max_rel_diff = 0.001
        if "m" in data_type:
            self.lam = 0.0001
            self.gam = 0.001
        else:
            self.lam = 0.01
            self.gam = 0.01
        self.lam_fac = 0.7
        self.lam_min = 1.E-6
        self.gam_fac = 0.7
        self.gam_min = 1.E-6
        self.sigma_mag = 1.
        self.sigma_grav = 1.
        self.sigma_sus = 1.E-6
        self.sigma_rem = 1.
        self.sigma_rho = 10.
        self.depth_ref = 1.
        self.width_max = 5
        self.max_amp = 0.1
        self.fig_sus = None
        self.fig_rem = None
        self.fig_rho = None
        self.fig_theo = None
        self.fig_theo2 = None
        self.pi4 = np.pi*4.
        self.param_prism = []
        self.par_hist = []
        self.stop_RMS_flag = False
        self.stop_Diff_flag = False
        self.RMS_misfit = []
        self.rel_RMS_misfit = []
        self.iteration = -1
        self.prism_del = {}
        self.prism_new = {}
        self.fig_follow = []
        now = datetime.now()
        self.c_time = now.strftime("%H-%M-%S")
        self.d1 = now.strftime("%Y-%m-%d")
        self.folder = f"inversion_{self.d1}_{self.c_time}"
        os.makedirs(self.folder)

    def run_inversion(self):
        """
        Do 2D or 3D inversion using a self-refining algorithm

        Returns
        -------
        None.

        """
        self.n_data = len(self.data)
        std_data = np.std(self.data)
        if self.iteration > 0:
            self.iteration -= 1
        i0 = 0
        if self.dim == 2:
            prism_add = 3
        else:
            prism_add = 7
# Start inversion
        while True:
            self.iteration += 1
            if self.max_iter == 0:
                print("\nOnly forward model calculated")
                self.G = self.mPrism.create_Frechet(
                    self.sus_inv, self.rem_inv, self.rho_inv, self.x,
                    self.y, self.z)
                self.data_mod = np.matmul(self.G, self.params)
                return
# Calculate effect of actual model
            print(f"\nStart iteration {self.iteration}")

# Create Frechet matrix and covariance matrices
            self.n_param = self.mPrism.get_n_param(
                self.sus_inv, self.rem_inv, self.rho_inv)
            self.S = self.mPrism.create_smooth(
                self.sus_inv, self.rem_inv, self.rho_inv, self.sigma_sus,
                self.sigma_rem, self.sigma_rho, self.depth_ref)
            self.G = self.mPrism.create_Frechet(
                self.sus_inv, self.rem_inv, self.rho_inv, self.x, self.y,
                self.z)
# Define data covariance and regularization matrices.
# Since both matrices have only values on their diagonal, values are stored as
# 1D vector. For the regularization matrix, the base value is the squared one
# given by the user interactively as parameter variability. This value is
# modified by the depth using as formula:
#    factor = (average_depth_of_prism/depth_ref) squared of gravity data and
#    to the power of three for magnetic data. Depth_ref is given interactively
#    by the user
            self.dat = np.copy(self.data)
            self.sigma_data, self.sigma_param = self.sigmas()
            print(f"Frechet calculated, shape: {self.G.shape}")

            if self.positive:
                fit = []
                self.dat = self.data_ori
                dd = np.copy(self.dat)
                self.params[:-1] = 0.001
                self.params[-1] = 0.
                for it in range(50):
                    self.yparam = np.copy(self.params)
                    self.yparam[:-1] = np.log(self.params[:-1])
                    fac = np.copy(self.params)
                    fac[-1] = 1.
                    G0 = self.G*fac
                    GCT = G0.T*self.sigma_data
                    G_inv = np.matmul(GCT, G0)
                    Gi_max = abs(G_inv).max()
                    Gp = self.lam*self.sigma_param*0.001
                    Gp_max = Gp.max()
                    Gp_fac = Gi_max/Gp_max
                    Gs = self.gam*self.S*0.001
                    Gs_max = Gs.max()
                    Gs_fac = Gi_max/Gs_max
                    if self.iteration == 0 and it == 0:
                        print(f"\nMaximum GT*Cd*G: {Gi_max:0.1f}")
                        print(f"   Maximum regularization: {Gp_max:0.1f} (fac:"
                              + f" {Gp_fac:0.1f})")
                        print(f"   Maximum smoothing     : {Gs_max:0.1f} (fac:"
                              + f" {Gs_fac:0.1f})")
                    G_inv += Gs
                    G_inv[np.diag_indices(G_inv.shape[0])] += Gp
                    d_par = np.matmul(np.matmul(np.linalg.inv(G_inv), GCT), dd)
                    yp = self.yparam+d_par
                    if yp[:-1].max() > 0.:
                        ip = np.argmax(yp[:-1])
                        factor = -self.yparam[ip]/d_par[ip]
                    else:
                        factor = 1.
                    self.yparam[:-1] += d_par[:-1]*factor
                    self.yparam[-1] += d_par[-1]
                    self.params = np.copy(self.yparam)
                    self.params[:-1] = np.exp(self.yparam[:-1])
                    self.params[:-1][self.params[:-1] < 1.E-7] = 1.E-7
                    self.data_mod = np.matmul(self.G, self.params)
                    dd = self.dat - self.data_mod
                    fit.append(np.std(dd))
                    if it < 2:
                        continue
                    if abs((fit[-2] - fit[-1])/fit[-2]) < 1.E-5:
                        break
# Do inversion
            else:
                GCT = self.G.T*self.sigma_data
                G_inv = np.matmul(GCT, self.G)
# For first iteration test whether regularization and smoothing matrices have
#     an appreciable effect. If not, give a warning message
                if self.iteration == 0:
                    mG = abs(G_inv).max()
                    print(f"\nMaximum GT*Cd*G: {mG:0.1f}")
                    mSig = (self.lam*self.sigma_param).max()
                    mSmo = (self.gam*self.S).max()
                    G_Sig = mG/mSig
                    G_Smo = mG/mSmo
                    print(f"   Maximum regularization: {mSig:0.1f} (fac: "
                          + f"{G_Sig:0.1f})")
                    print(f"   Maximum smoothing: {mSmo:0.1f} (fac: "
                          + f"{G_Smo:0.1f})")
                    if G_Sig > 10 or G_Smo > 10:
                        answer = QtWidgets.QMessageBox.warning(
                            None, "Warning",
                            "Regularization or smoothing much smaller than "
                            + "Frechet\n"
                            + f"     Frechet/Regularization: {G_Sig:0.0f}\n"
                            + f"     Frechet/Smoothing: {G_Smo:0.0f}\n"
                            + "You may increase lambda/gamma or restart and "
                            + "decrease parameter variances\n\n"
                            + "Ignore to continue nevertheless\nRetry to give "
                            + "new initial lambda/gamma\nAbort "
                            + "to stop inversion\n",
                            QtWidgets.QMessageBox.Ignore |
                            QtWidgets.QMessageBox.Abort |
                            QtWidgets.QMessageBox.Retry,
                            QtWidgets.QMessageBox.Ignore)
                        if answer == QtWidgets.QMessageBox.Abort:
                            sys.exit()
                        elif answer == QtWidgets.QMessageBox.Retry:
                            results, okButton = dialog(
                                    ["New Initial lambda (regularization)",
                                     "New initial gamma (smoothing)"],
                                    ["e", "e"], [self.lam, self.gam],
                                    "New inversion parameters")
                            if okButton:
                                self.lam = float(results[0])
                                self.gam = float(results[1])

                G_inv[np.diag_indices(G_inv.shape[0])] +=\
                    self.lam*self.sigma_param
                G_inv += self.gam*self.S
                d_par = np.matmul(np.matmul(np.linalg.inv(G_inv), GCT),
                                  self.dat)
                self.params += d_par
            self.data_mod = np.matmul(self.G, self.params)
# Extract new prism properties from parameter vector, set the correspondig
#   values in the prism parameters and copy them into vector par_hist
            i0 = 0
            ss = []
            if self.sus_inv:
                for i, key in enumerate(self.mPrism.prisms.keys()):
                    self.mPrism.prisms[key].sus = self.params[i]
                    ss.append(self.mPrism.prisms[key].sus)
                i0 += self.mPrism.n_prisms
            if self.rem_inv:
                sr = []
                for i, key in enumerate(self.mPrism.prisms.keys()):
                    self.mPrism.prisms[key].rem = self.params[i+i0]
                    sr.append(self.mPrism.prisms[key].rem)
            if self.sus_inv:
                if self.rem_inv:
                    for i, s in enumerate(ss):
                        self.par_hist.append(np.array([s*self.pi4, sr[i]]))
                else:
                    self.par_hist.append(np.array(ss)*self.pi4)
            elif self.rem_inv:
                self.par_hist.append(sr)
            if self.rho_inv:
                s = []
                for i, key in enumerate(self.mPrism.prisms.keys()):
                    self.mPrism.prisms[key].rho = self.params[i]
                    s.append(self.mPrism.prisms[key].rho)
                self.par_hist.append(np.array(s))
# Calculate magnetic misfits and some statistics
            data = self.data_ori-self.data_mod
            data -= np.nanmedian(data)
            self.data = np.copy(data)
            std_data0 = std_data
            std_data = np.std(data)
            self.std_data_rel = std_data/self.std_data_ori
            std_data_diff = (std_data0-std_data)/self.std_data_ori
            self.RMS_misfit.append(std_data)
            self.rel_RMS_misfit.append(self.std_data_rel*100.)
            if self.std_data_rel < self.max_diff_fac:
                self.stop_RMS_flag = True
            if abs(std_data_diff) < self.max_rel_diff:
                self.stop_Diff_flag = True
            print(f"Maximum data difference data: {abs(data).max()}")
            print(f"Relative Std misfit [%]: {self.std_data_rel*100}")
            print(f"Relative variation of misfit [%]: {std_data_diff*100}")
            print(f"Average: {self.params[-1]}, "
                  + f"sus:({self.params[:-1].min()*1000.}, "
                  + f"{self.params[:-1].max()*1000.})\n")
# If Maximum iteration number is reached, misfit does not become smaller or
# relative misfits are smaller than predefined minimum value, stop iterations
            if self.iteration == self.max_iter:
                print("\nMaximum number of iterations reached\n"
                      + "Iterations stop")
                return
            if self.stop_RMS_flag:
                print("\nMisfit limit reached\nIterations stop")
                return
            if self.stop_Diff_flag:
                print("\nNo more misfit reduction.")
# Modify prism size in area of maximum misfit
            key_m = []
            key_r = []
            self.stop_Diff_flag = False
# In the following lines, only the mPrism.get_max_prisms lines are important.
# The lineswith utils.get_extremes are only for testing purposes, such as the
# plots later on.
            if self.n_sensor == 2:
                _, maxima1 = utils.get_extremes(
                    abs(data[:self.n_data1]).
                    reshape(self.data1_shape), self.width_max)
                _, maxima2 = utils.get_extremes(
                    abs(data[self.n_data1:]).
                    reshape(self.data2_shape), self.width_max)
                maxima = maxima1+maxima2
            else:
                _, maxima = utils.get_extremes(
                    abs(data).reshape(self.data_shape), self.width_max)
            for pos in maxima:
                if len(pos) == 1:
                    i = pos[0]
                else:
                    i = pos[0]*self.data_shape[1] + pos[1]
            key_r = []
            if self.n_sensor == 2:
                key_m1 = self.mPrism.get_max_prisms(
                    abs(data[:self.n_data1]).
                    reshape(self.data1_shape),
                    self.G[:self.n_data1, :self.n_prisms],
                    max_lim=self.max_amp, width=self.width_max)
                key_m2 = self.mPrism.get_max_prisms(
                    abs(data[self.n_data1:]).
                    reshape(self.data2_shape),
                    self.G[self.n_data1:, :self.n_prisms],
                    max_lim=self.max_amp, width=self.width_max)
                key_m = list(np.unique(np.array(key_m1+key_m2)))
            else:
                key_m = self.mPrism.get_max_prisms(
                    abs(data.reshape(self.data_shape)),
                    self.G[:self.n_data1, :self.n_prisms],
                    width=self.width_max)
            if self.sus_inv and self.rem_inv:
                if self.n_sensor == 2:
                    key_r1 = self.mPrism.get_max_prisms(
                        abs(data[:self.n_data1]).
                        reshape(self.data1_shape),
                        self.G[:self.n_data1, self.n_prisms:],
                        max_lim=self.max_amp, width=self.width_max)
                    key_r2 = self.mPrism.get_max_prisms(
                        abs(data[self.n_data1:]).
                        reshape(self.data2_shape),
                        self.G[self.n_data1:, self.n_prisms:],
                        max_lim=self.max_amp, width=self.width_max)
                    key_r = list(np.unique(np.array(key_r1+key_r2)))
                else:
                    key_r = self.mPrism.get_max_prisms(
                        abs(data.reshape(self.data1_shape)),
                        self.G[:self.n_data1, self.n_prisms:],
                        max_lim=self.max_amp, width=self.width_max)
            key_split = list(np.unique(np.array(key_m+key_r)))
# Test whether prisms are marked for splitting
            if len(key_split) == 0:
                print("  Prisms reached size limit\nIteration stops")
                return
# Test whether number of prisms becomes larger than number of data points after
# splitting Prisms will be split into up to 8 pieces, the prism itself is
# eliminated. Therefore the factor 7 in the next line.
            if self.n_prisms+prism_add*len(key_split) > self.n_data:
                print("Inversion stopped: More prisms than data points")
                return
            self.prism_del[self.iteration] = {}
            self.prism_new[self.iteration] = {}
            print(f"  Split prisms {key_split}")
            for _, key in enumerate(key_split):
                self.prism_del[self.iteration][key] =\
                    deepcopy(self.mPrism.prisms[key])
                key_add = self.mPrism.split(key)
                self.prism_new[self.iteration][key] = {}
                for k in key_add:
                    self.prism_new[self.iteration][key][k] =\
                        deepcopy(self.mPrism.prisms[k])
            self.n_prisms = len(list(self.mPrism.prisms.keys()))
            self.params = []
            param_prism = []
            if self.sus_inv:
                for key, val in self.mPrism.prisms.items():
                    self.params.append(val.sus)
                    param_prism.append(key)
            if self.rem_inv:
                for key, val in self.mPrism.prisms.items():
                    self.params.append(val.rem)
                    param_prism.append(key)
            if self.rho_inv:
                for key, val in self.mPrism.prisms.items():
                    self.params.append(val.rho)
                    param_prism.append(key)
            self.params.append(0.)
            self.params = np.array(self.params)
            self.lam = max(self.lam*self.lam_fac, self.lam_min)
            self.gam = max(self.gam*self.gam_fac, self.gam_min)
            print(f"New lambda: {self.lam}; new gamma: {self.gam}")
            if self.max_iter == 0:
                return

    def sigmas(self):
        icol = 0
        if self.depth_ref > 1.:
            fac = (self.depth_ref-1.)/self.zprism_max
        elif self.depth_ref < 1.:
            fac = (1./self.depth_ref-1.)/self.zprism_max
        else:
            fac = 1.
        if "m" in self.data_type:
            sigma_data = np.ones(self.n_data)/self.sigma_mag**2
            i = -1
            if self.sus_inv:
                sigma_param = np.ones(self.mPrism.n_prisms)\
                    / self.sigma_sus**2
                if not np.isclose(self.depth_ref, 1.):
                    for _, val in self.mPrism.prisms.items():
                        i += 1
                        f = 1.+(val.z[0]+val.z[1])/2.*fac
                        if self.depth_ref < 1.:
                            f = 1./f
                        sigma_param[i] *= f**3
                icol += self.n_prisms
            elif self.rem_inv:
                sigma_param = np.ones(self.mPrism.n_prisms)\
                    / self.sigma_rem**2
                if not np.isclose(self.depth_ref, 1.):
                    for _, val in self.mPrism.prisms.items():
                        i += 1
                        f = 1.+(val.z[0]+val.z[1])/2.*fac
                        if self.depth_ref < 1.:
                            f = 1./f
                        sigma_param[i] *= f**3
                icol += self.n_prisms
        else:
            sigma_data = np.ones(self.n_data)/self.sigma_grav**2
            sigma_param =\
                np.ones(self.mPrism.n_prisms)/self.sigma_rho**2
            i = -1
            if not np.isclose(self.depth_ref, 1.):
                for _, val in self.mPrism.prisms.items():
                    i += 1
                    f = 1.+(val.z[0]+val.z[1])/2.*fac
                    if self.depth_ref < 1.:
                        f = 1./f
                    sigma_param[i] *= f**2
        sigma_param = np.concatenate((sigma_param, np.array([0.])))
        return sigma_data, sigma_param

    def show_results2D(self, file):
        """
        Plot results of 2D inversion.

        Parameters
        ----------
        file : str
            Base name of file where to store the image

        """
        equal_flag = False
        cmap = plt.get_cmap("rainbow")
        if self.sus_inv:
            txt = "Susceptibility"
            title = "Magnetic data"
            ylabel = "Anomaly [nT]"
            c_txt = "Sus [SI*10**-3]"
        elif self.rem_inv:
            txt = "Remanence"
            title = "Magnetic data: "
            ylabel = "Anomaly [nT]"
            c_txt = "Remanence [A/m]"
        else:
            txt = "Density"
            title = "Gravity data"
            ylabel = "Anomaly [mGal]"
            c_txt = "Density [kg/m3]"
        if self.direction in ("N", "S", "Y", 0., 180.):
            title += f" at X={self.line_pos*self.dfac}"
        else:
            title += f" at Y={self.line_pos*self.dfac}"
        title += f"; RMS misfit: {self.std_data_rel*100:0.2f}%"
        if file[-1] == "N":
            sens = "Easting"
        else:
            sens = "Northing"
        while True:
            self.fig_inv2 = newWindow(f"{txt} model", 2000, 1500, 15, 15)
            self.fig_inv2.fig.tight_layout(w_pad=15, h_pad=2)
            self.gs = GridSpec(16, 10, self.fig_inv2.fig)
            ax_dat = self.fig_inv2.fig.add_subplot(self.gs[1:7, 1:9])
            ax_mod = self.fig_inv2.fig.add_subplot(self.gs[10:16, 1:9])
            xmin = self.x.min()
            xmax = self.x.max()
            if xmax-xmin > 10000:
                xx = self.x/1000.
                xmin /= 1000.
                xmax /= 1000.
                unit = "km"
                fac = 0.001
            else:
                xx = self.x
                unit = "m"
                fac = 1.
            if self.n_sensor == 1:
                dat = self.data_ori+np.nanmedian(self.data_mod)
                ax_dat.plot(xx, dat, "b*", label="meas. data")
                ax_dat.plot(xx, self.data_mod, "cyan", label="calc. data")
            else:
                dat = self.data_ori[:self.n_data1]\
                    + np.nanmedian(self.data_mod[:self.n_data1])\
                    - np.nanmedian(self.data_ori[:self.n_data1])
                ax_dat.plot(xx[:self.n_data1], dat, "b*",
                            label="meas. data sensor 1")
                ax_dat.plot(xx[:self.n_data1], self.data_mod[:self.n_data1],
                            "cyan", label="calc. data sensor 1")
                dat = self.data_ori[self.n_data1:]\
                    + np.nanmedian(self.data_mod[self.n_data1:])\
                    - np.nanmedian(self.data_ori[self.n_data1:])
                ax_dat.plot(xx[self.n_data1:], dat, "r+",
                            label="meas. data sensor 2")
                ax_dat.plot(xx[self.n_data1:], self.data_mod[self.n_data1:],
                            "orange", label="calc. data sensor 2")
                ax_dat.legend(bbox_to_anchor=(1, 1), loc="upper right",
                              fontsize=10)
            ax_dat.tick_params(axis='both', labelsize=12)
            ax_dat.set_title(title, fontsize=14)
            ax_dat.set_xlabel(f"{sens} [{unit}]", fontsize=12)
            ax_dat.set_ylabel(ylabel, fontsize=12)
            ax_dat.set_xlim([xmin, xmax])
            cax = ax_dat.inset_axes([1.05, 0.05, 0.02, 0.9],
                                    transform=ax_dat.transAxes)
            cax.axis("off")
            patches = []
            col = []
            ymin = 1e10
            ymax = -1E10
            for key, val in self.mPrism.prisms.items():
                x1 = val.x[0]*fac
                x2 = val.x[1]*fac
                y1 = val.z[0]*fac
                y2 = val.z[1]*fac
                ymin = min(ymin, y1)
                ymax = max(ymax, y2)
                if self.sus_inv:
                    col.append(val.sus*self.pi4*1000.)
                elif self.rem_inv:
                    col.append(val.rem)
                else:
                    col.append(val.rho)
                patches.append(Rectangle((x1, y1), x2-x1, y2-y1))
            col = np.array(col)
            p = PatchCollection(patches, cmap=cmap)
            p.set_array(col)
            ax_mod.add_collection(p)
            ax_mod.set_xlim([xmin, xmax])
            ax_mod.set_ylim([ymin, ymax])
            ax_mod.invert_yaxis()
            ax_mod.tick_params(axis='both', labelsize=12)
            ax_mod.set_title(txt, fontsize=14)
            ax_mod.set_xlabel(f"{sens} [{unit}]", fontsize=12)
            ax_mod.set_ylabel(f"Depth [{unit}]", fontsize=12)
            if equal_flag:
                ax_mod.set_aspect("equal")
            else:
                ax_mod.set_aspect("auto")
            cax = ax_mod.inset_axes([1.05, 0.05, 0.02, 0.9],
                                    transform=ax_mod.transAxes)
# Plot color bar
            cbar = plt.colorbar(p, orientation="vertical", cax=cax,
                                fraction=0.1)
            cbar.set_label(label=c_txt, size=12)
            for lab in cbar.ax.yaxis.get_ticklabels():
                lab.set_fontsize(12)
            self.fig_inv2.setHelp(
                "Press ENTER to finish; press r to continue iterations; "
                + "press c to toggle between equal scale and filling window "
                + "for model axis. If no reaction, move mouse a little bit")
            self.fig_inv2.show()
# Save inversion control parameters to file parameters.dat
            file_name = os.path.join(self.folder, "parameters.dat")
            self.write_parameters(file_name)
            if equal_flag:
                fil = os.path.join(self.folder, file+"_scaled.png")
            else:
                fil = os.path.join(self.folder, file+"_not-scaled.png")
            self.fig_inv2.fig.savefig(fil)
            while True:
                event = self.fig_inv2.get_event()
                self.fig_inv2.close()
                if event.name == "key_press_event":
                    if event.key == "enter":
                        return False
                    if event.key in ("c", "C"):
                        equal_flag = not equal_flag
                        break
                    elif event.key in ("r", "R"):
                        return True

    def show_results3D(self):
        """
        Plot inversion results
        - maps of prism parameters (susceptibilities, remanences of densities)
        - maps of calculated data and difference calculated minus measured
        - evolution of misfit

        Returns
        -------
        None.

        """
        self.set_plot_depths()
        if self.sus_inv:
            txt = "Susceptibility"
            c_txt = "Sus [SI*10**-3]"
        elif self.rem_inv:
            txt = "Remanence"
            c_txt = "Remanence [A/m]"
        else:
            txt = "Density"
            c_txt = "Density [kg/m3]"
# Calculate number of subplots to be done, depending on number of layers
#    calculated
        i0 = 0
        n_prop_plots = len(self.nz_plot)
        max_ax_col = int(np.ceil(np.sqrt(n_prop_plots)))
        max_ax_row = int(np.ceil(n_prop_plots/max_ax_col))
        nax_plot = n_prop_plots-max_ax_col
        npltx = min(max_ax_col, n_prop_plots)
        nplty = max_ax_row
        width = npltx*5
        height = nplty*5
        par = []
# Plot model parameter distribution
        for key, val in self.mPrism.prisms.items():
            if self.sus_inv:
                par.append(val.sus*self.pi4*1000.)
            elif self.rem_inv:
                par.append(val.rem)
            else:
                par.append(val.rho)
        par = np.array(par)
        i0 += self.mPrism.n_prisms
        k = 0
        vmin = np.quantile(par, 0.01)
        vmax = np.quantile(par, 0.99)
        norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)
        self.fig_par = newWindow(f"{txt} model", 2400, 1800, width, height)
        self.fig_par.fig.tight_layout(w_pad=15, h_pad=2)
        self.gs = GridSpec(nplty*10, npltx*10+2, self.fig_par.fig)
        figy0 = -10
        for i in range(nplty):
            figy0 += 10
            figx0 = -10
            for j in range(npltx):
                figx0 += 10
                patches = []
                col = []
                ax = self.fig_par.fig.add_subplot(self.gs[figy0:figy0+8,
                                                          figx0:figx0+8])
                if k >= n_prop_plots:
                    ax.axis("off")
                    k += 1
                    continue
                for key, val in self.mPrism.prisms.items():
                    if val.z[0] <= self.z_plot[k] <= val.z[1]:
                        if self.xprism_min > 100000.:
                            x1 = val.x[0]/1000.
                            x2 = val.x[1]/1000.
                            y1 = val.y[0]/1000.
                            y2 = val.y[1]/1000.
                        else:
                            x1 = val.x[0]
                            x2 = val.x[1]
                            y1 = val.y[0]
                            y2 = val.y[1]
                        if self.sus_inv:
                            col.append(val.sus*self.pi4*1000.)
                        elif self.rem_inv:
                            col.append(val.rem)
                        else:
                            col.append(val.rho)
                        patches.append(Rectangle((x1, y1), x2-x1, y2-y1))
                col = np.array(col)
                p = PatchCollection(patches, cmap="rainbow", norm=norm,
                                    edgecolors=("black",))
                p.set_array(col)
                ax.add_collection(p)
                ax.set_title(f"{txt} at {self.z_plot[k]:0.1f} m", fontsize=10)
                ax.tick_params(top=True, right=True, labelsize=10)
                if k+npltx+1 > n_prop_plots:
                    ax.set_xlabel(f"Easting [{self.ax_unit}]", fontsize=10)
                else:
                    ax.set_xticklabels([])
                if j == 0:
                    ax.set_ylabel(f"Northing [{self.ax_unit}]", fontsize=10)
                elif j == npltx-1 or k == n_prop_plots-1:
                    ax.yaxis.tick_right()
                else:
                    ax.set_yticklabels([])
                ax.set_xlim([self.xprism_min, self.xprism_max])
                ax.set_ylim([self.yprism_min, self.yprism_max])
                ax.set_aspect("equal", adjustable="box")
                if k < nax_plot:
                    ax.set_xlabel("")
                if j > 0:
                    ax.set_ylabel("")
                k += 1
        ax = self.fig_par.fig.add_subplot(self.gs[1:-1, -1:])
        ax.axis("off")
        cax = ax.inset_axes([0., 0., 0.9, 0.9], transform=ax.transAxes)
        cbar = plt.colorbar(p, orientation="vertical", cax=cax, fraction=0.9)
        cbar.set_label(c_txt, fontsize=10)
        cbar.ax.tick_params(labelsize=10)

        self.fig_par.setHelp("Press ENTER to finish, R to continue iterations")
        self.fig_par.show()
        self.write_parameters(os.path.join(self.folder, "parameters.dat"))
        self.fig_par.fig.savefig(
            os.path.join(self.folder, f"{txt}_distribution.png"))

        if "m" in self.data_type:
            self.fig_theo = newWindow("Magnetic data", 1500, 1000, 20, 15)
            self.fig_theo.fig.tight_layout(w_pad=15, h_pad=2)
            plt_file = os.path.join(self.folder, "Mag_sensor1_calc&diff.png")
            title = "Modelled magnetic sensor1"
            unit = "nT"
            title2 = "Mag difference sensor1\nrel. RMS misfit: "\
                + f"{self.std_data_rel*100:0.2f}%"
        else:
            self.fig_theo = newWindow("Gravity data", 1500, 1000, 20, 15)
            self.fig_theo.fig.tight_layout(w_pad=15, h_pad=2)
            plt_file = os.path.join(self.folder, "Gravi_calc&diff.png")
            title = "Modelled gravity data"
            unit = "mGal"
            title2 = "Gravity difference\nrel. RMS misfit: "\
                + f"{self.std_data_rel*100:0.2f}%"
        self.ax_theo = []
        ddx = self.xplt_max - self.xplt_min
        ddy = self.yplt_max - self.yplt_min
        if ddx > 1.5*ddy:
            self.gs = GridSpec(18, 10, self.fig_theo)
            self.ax_theo.append(
                self.fig_theo.fig.add_subplot(self.gs[1:8, 1:]))
            self.ax_theo.append(
                self.fig_theo.fig.add_subplot(self.gs[10:17, 1:]))
            self.bar_or = "vertical"
#                    anchor = 'E'
            self.nticks = 10
# Horizontal layout
        else:
            self.gs = GridSpec(10, 18, self.fig_theo)
            self.ax_theo.append(
                self.fig_theo.fig.add_subplot(self.gs[1:, 1:8]))
            self.ax_theo.append(
                self.fig_theo.fig.add_subplot(self.gs[1:, 10:17]))
            self.bar_or = "horizontal"
#                    anchor = 'S'
            self.nticks = 5

# Plot magnetic anomalies produced by inverted model
        data = self.data_mod[:self.n_data1].reshape(self.data1_shape)
        med = np.nanmedian(data)
        data -= med
        vmin = np.ceil(np.quantile(data, 0.005)*1000)/1000
        vmax = np.ceil(np.quantile(data, 0.995)*1000)/1000
        br_map, norm = utils.mag_color_map(vmin, vmax)
        im, cbar = utils.data_plot(
            data, self.fig_theo.fig, self.ax_theo[0],
            title=f"{title}\nMedian: {med:0.1f}",
            xtitle=f"Easting [{self.ax_unit}]",
            ytitle=f"Northing [{self.ax_unit}]",
            cmap=br_map, norm=norm, cbar_title=unit,
            extent=[self.xplt_min, self.xplt_max,
                    self.yplt_min, self.yplt_max])
# plot prism contours
        for key, val in self.mPrism.prisms.items():
            x1 = val.x[0]
            x2 = val.x[1]
            y1 = val.y[0]
            y2 = val.y[1]
            self.ax_theo[0].plot([x1, x2, x2, x1, x1], [y1, y1, y2, y2, y1],
                                 "k", linewidth=1)
        self.ax_theo[0].set_xlim([self.xplt_min, self.xplt_max])
        self.ax_theo[0].set_ylim([self.yplt_min, self.yplt_max])
        self.ax_theo[0].grid(visible=True, which="both")
        self.ax_theo[0].set_xlabel("")
# Plot misfit
        data = (self.data_mod[:self.n_data1] -
                self.data_ori[:self.n_data1]).reshape(self.data1_shape)
        data -= np.nanmedian(data)
        vmin = np.ceil(np.quantile(data, 0.005)*1000)/1000
        vmax = np.floor(np.quantile(data, 0.995)*1000)/1000
        br_map, norm = utils.mag_color_map(vmin, vmax)
        im, cbar = utils.data_plot(
            data, self.fig_theo.fig, self.ax_theo[1], title=f"{title2}",
            xtitle=f"Easting [{self.ax_unit}]",
            ytitle=f"Northing [{self.ax_unit}]",
            extent=[self.xplt_min, self.xplt_max,
                    self.yplt_min, self.yplt_max], cmap=br_map,
            cbar_title=f"Difference calc-meas [{unit}]", norm=norm)
        self.ax_theo[1].grid(visible=True, which="both")
        self.fig_theo.show()
        self.fig_theo.fig.savefig(plt_file)

# Plot results for sensor 2
        if self.n_sensor == 2:
            self.fig_theo2 = newWindow("Magnetic data", 1500, 1000, 24, 18)
            self.fig_theo2.fig.tight_layout(w_pad=15, h_pad=2)
            plt_file2 = os.path.join(self.folder, "Mag_sensor2_calc&diff.png")
            title2 = "Mag difference sensor2, rel. RMS misfit: "\
                + f"{self.std_data_rel*100:0.2f}%"
            self.ax_theo2 = []
            if ddx > 1.5*ddy:
                self.gs = GridSpec(18, 10, self.fig_theo2)
                self.ax_theo2.append(
                    self.fig_theo2.fig.add_subplot(self.gs[1:8, 1:]))
                self.ax_theo2.append(
                    self.fig_theo2.fig.add_subplot(self.gs[10:17, 1:]))
                self.bar_or = "vertical"
#                    anchor = 'E'
                self.nticks = 10
# Horizontal layout
            else:
                self.gs = GridSpec(10, 18, self.fig_theo2)
                self.ax_theo2.append(
                    self.fig_theo2.fig.add_subplot(self.gs[1:, 1:8]))
                self.ax_theo2.append(
                    self.fig_theo2.fig.add_subplot(self.gs[1:, 10:17]))
                self.bar_or = "horizontal"
#                    anchor = 'S'
                self.nticks = 5
# Plot magnetic anomalies produced by inverted model
            data = self.data_mod[self.n_data1:].reshape(self.data2_shape)
            med = np.median(data)
            data -= med
            vmin = np.ceil(np.quantile(data, 0.005)*1000)/1000
            vmax = np.ceil(np.quantile(data, 0.995)*1000)/1000
            br_map, norm = utils.mag_color_map(vmin, vmax)
            im, cbar = utils.data_plot(
                data, self.fig_theo2.fig, self.ax_theo2[0],
                title=f"Modelled magnetic sensor2\nMedian: {med:0.1f}",
                xtitle=f"Easting [{self.ax_unit}]",
                ytitle=f"Northing [{self.ax_unit}]",
                cmap=br_map, norm=norm, cbar_title=unit,
                extent=[self.xplt_min, self.xplt_max,
                        self.yplt_min, self.yplt_max])
# plot prism contours
            for key, val in self.mPrism.prisms.items():
                x1 = val.x[0]
                x2 = val.x[1]
                y1 = val.y[0]
                y2 = val.y[1]
                self.ax_theo2[0].plot([x1, x2, x2, x1, x1],
                                      [y1, y1, y2, y2, y1], "k", linewidth=1)
            self.ax_theo2[0].set_xlim([self.xplt_min, self.xplt_max])
            self.ax_theo2[0].set_ylim([self.yplt_min, self.yplt_max])
            self.ax_theo2[0].grid(visible=True, which="both")
            self.ax_theo2[0].set_xlabel("")

# Plot misfit
            data = (self.data_mod[self.n_data1:] -
                    self.data_ori[self.n_data1:]).\
                reshape(self.data2_shape)
            data -= np.nanmedian(data)
            vmin = np.ceil(np.quantile(data, 0.005)*1000)/1000
            vmax = np.floor(np.quantile(data, 0.995)*1000)/1000
            br_map, norm = utils.mag_color_map(vmin, vmax)
            im, cbar = utils.data_plot(
                data, self.fig_theo2.fig, self.ax_theo2[1],
                title="Mag difference sensor2\n"
                + f"rel. RMS misfit: {self.std_data_rel*100:0.2f}%",
                xtitle=f"Easting [{self.ax_unit}]",
                ytitle=f"Northing [{self.ax_unit}]",
                extent=[self.xplt_min, self.xplt_max,
                        self.yplt_min, self.yplt_max],
                cbar_title="Difference calc-meas [nT]", cmap=br_map, norm=norm)
            self.ax_theo2[1].grid(visible=True, which="both")
            self.fig_theo2.show()
            self.fig_theo2.fig.savefig(plt_file2)

# Plot misfit evolution
        self.fig_misfit = newWindow("Magnetic data", 800, 500)
        self.ax_misfit = self.fig_misfit.fig.subplots(1, 1)
        self.ax_misfit.plot(np.arange(len(self.rel_RMS_misfit)),
                            self.rel_RMS_misfit, "k")
        self.ax_misfit.set_title("Misfit evolution")
        self.ax_misfit.set_xlabel("Iteration number")
        self.ax_misfit.set_ylabel("RMS misfit [%]")
        self.fig_misfit.show()
        self.fig_misfit.fig.savefig(os.path.join(self.folder,
                                                 "misfit-evolution.png"))
        print("\nClick into parameter distribution window to finish inversion")
        while True:
            event = self.fig_par.get_event()
            if event.name == "key_press_event":
                if event.key in ("enter", "r", "R"):
                    break
        self.fig_par.close()
        self.fig_theo.close()
        self.fig_misfit.close()
        if self.n_sensor == 2:
            self.fig_theo2.close()
        if event.key == "enter":
            return False
        else:
            return True

    def show_synthetic(self):
        """
        Plots results from synthetic model calculation

        """
        if "m" in self.data_type:
            self.fig_syn = newWindow("Magnetic data", 1500, 1000, 20, 15)
            self.fig_syn.fig.tight_layout(w_pad=15, h_pad=2)
            plt_file = "Mag_synthetic.png"
            title = f"Modelled magnetic anomaly {self.z[0]}m"
            if self.n_sensor == 2:
                title2 = f"Modelled magnetic anomaly {self.z2[0]}m"
            unit = "nT"
        else:
            self.fig_syn = newWindow("Gravity data", 1500, 1000, 20, 15)
            self.fig_syn.fig.tight_layout(w_pad=15, h_pad=2)
            plt_file = "Gravi_synthetic.png"
            title = "Modelled gravity anomaly"
            unit = "mGal"
# Plot magnetic anomalies produced by synthetic model
        data = self.data_mod[:self.n_data1].reshape(self.data1_shape)
        if self.n_sensor == 2:
            data2 = self.data_mod[self.n_data1:].reshape(self.data2_shape)
            ddx = self.x.max() - self.x.min()
            ddy = self.y.max() - self.y.min()
            facx = 10 / (2 * ddx)
            facy = 8 / (2 * ddy)
# Vertical layout
            self.ax_syn = []
            if facx < facy:
                self.gs = GridSpec(18, 10, self.fig_syn.fig)
                self.ax_syn.append(self.fig_syn.fig.add_subplot(
                    self.gs[1:8, 1:]))
                self.ax_syn.append(self.fig_syn.fig.add_subplot(
                    self.gs[10:17, 1:]))
                self.bar_or = "vertical"
                self.nticks = 10
# Horizontal layout
            else:
                self.gs = GridSpec(10, 18, self.fig_syn.fig)
                self.ax_syn.append(self.fig_syn.fig.add_subplot(
                    self.gs[1:, 1:8]))
                self.ax_syn.append(self.fig_syn.fig.add_subplot(
                    self.gs[1:, 10:17]))
                self.bar_or = "horizontal"
                self.nticks = 5
        else:
            self.gs = GridSpec(10, 10, self.fig_syn.fig)
            self.ax_syn = [self.fig_syn.fig.add_subplot(self.gs[1:-1, 1:-1])]
        vmin = np.ceil(np.quantile(data, 0.005)*1000)/1000
        vmax = np.ceil(np.quantile(data, 0.995)*1000)/1000
        self.xplt_min = self.x.min()
        self.xplt_max = self.x.max()
        self.yplt_min = self.y.min()
        self.yplt_max = self.y.max()
        if "m" in self.data_type:
            br_map, norm = utils.mag_color_map(vmin, vmax)
            im, cbar = utils.data_plot(
                data, self.fig_syn.fig, self.ax_syn[0], title=f"{title}",
                xtitle="Easting [m]",
                ytitle="Northing [m]",
                cmap=br_map, norm=norm, cbar_title=unit,
                extent=[self.xplt_min, self.xplt_max,
                        self.yplt_min, self.yplt_max])
            if self.n_sensor == 2:
                im, cbar = utils.data_plot(
                    data2, self.fig_syn.fig, self.ax_syn[1], title=f"{title2}",
                    xtitle="Easting [m]",
                    ytitle="Northing [m]",
                    cmap=br_map, norm=norm, cbar_title=unit,
                    extent=[self.xplt_min, self.xplt_max,
                            self.yplt_min, self.yplt_max])
        else:
            im, cbar = utils.data_plot(
                data, self.fig_syn.fig, self.ax_syn, title=f"{title}",
                xtitle="Easting [m]",
                ytitle="Northing [m]",
                cmap="rainbow", cbar_title=unit,
                extent=[self.xplt_min, self.xplt_max,
                        self.yplt_min, self.yplt_max])
# plot prism contours
        for key, val in self.mPrism.prisms.items():
            x1 = val.x[0]
            x2 = val.x[1]
            y1 = val.y[0]
            y2 = val.y[1]
            self.ax_syn[0].plot([x1, x2, x2, x1, x1], [y1, y1, y2, y2, y1],
                                "k", linewidth=1)
            if self.n_sensor == 2:
                self.ax_syn[1].plot([x1, x2, x2, x1, x1], [y1, y1, y2, y2, y1],
                                    "k", linewidth=1)
        self.ax_syn[0].set_xlim([self.xplt_min, self.xplt_max])
        self.ax_syn[0].set_ylim([self.yplt_min, self.yplt_max])
        self.ax_syn[0].grid(visible=True, which="both")
        if self.n_sensor == 2:
            self.ax_syn[1].set_xlim([self.xplt_min, self.xplt_max])
            self.ax_syn[1].set_ylim([self.yplt_min, self.yplt_max])
            self.ax_syn[1].grid(visible=True, which="both")
        self.fig_syn.fig.savefig(plt_file)
        self.fig_syn.show()
        print("\nClick to close window and finish synthetic calculation")
        while True:
            event = self.fig_syn.get_event()
            if event.name == "button_press_event":
                break
        self.fig_syn.close()

    def save_model(self):
        """
        Write model parameters into file model_date&time.dat

        Parameters
        ----------
        file_name : str
            Name of ASCII file where to store the model parameters
        """
        with open(os.path.join(self.folder, "model.dat"), "w") as fo:
            text1 = ""
            text2 = ""
            if self.sus_inv:
                text1 += "     sus    "
                text2 += "  [1E-6 SI] "
            if self.rem_inv:
                text1 += "  rem  "
                text2 += " [A/m] "
            if self.rho_inv:
                text1 += "  rho  "
                text2 += "[km/m3]"
            fo.write(f"           Prism_center      {text1}                   "
                     + " Prism coordinates\n")
            fo.write(f"      X          Y       Z   {text2}   #       W       "
                     + "  E          S          N      top   bottom\n")
            for key, val in self.mPrism.prisms.items():
                x1 = val.x[0]
                x2 = val.x[1]
                xc = (x1+x2)*0.5
                y1 = val.y[0]
                y2 = val.y[1]
                yc = (y1+y2)*0.5
                z1 = val.z[0]
                z2 = val.z[1]
                zc = (z1+z2)*0.5
                fo.write(f"{xc:9.2f} {yc:10.2f} {zc:7.2f} ")
                if self.sus_inv:
                    fo.write(f"{val.sus*4*np.pi*1E6:10.0f}")
                if self.rem_inv:
                    fo.write(f"{val.rem:7.3f}")
                if self.rho_inv:
                    fo.write(f"{val.rho:7.0f}")
                fo.write(f"{key:6d} {x1:9.2f} {x2:9.2f} {y1:10.2f} {y2:10.2f} "
                         + f"{z1:7.2f} {z2:7.2f}\n")
        return None

    def get_inversion_parameters(self, data_type):
        """
        Ask for inversion control parameters

        Parameters
        ----------
        data_type : str
            Data to be inverted: "m1" or "m2" for magnetic data,
            "g" for gravity data

        Defines the following variables
        -------------------------------
        sus_inv, rem_inv, rho_inv : booleans
            If True, susceptibility, remanence or density inversion activated
        max_iter : int
            Maximum number of iterations
        max_diff_fac : float
            Iterations stop if maximum relative RMS fit of each data set is
            smaller than the given value [%/100]
        max_rel_diff : float
            Iterations stop if variation of maximum relative RMS fit of each
            data set from one iteration to the next is smaller than the given
            value [%/100]
        lam, lam_fac, lam_min: floats
            initial lambda, factor per iteration step, smallest allowed lambda
        gam, gam_fac, gam_min: floats
            initial gamma, factor per iteration step, smallest allowed gamma

        """
        results = False
        if "m" in data_type:
            results, okButton = dialog(
                    ["Invert for:",
                     "Susceptibility",
                     "Remanence",
                     "Maximum number of iterations",
                     "Maximum relative RMS misfit",
                     "Maximum variation of RMS",
                     "Initial lambda (regularization)",
                     "Lambda factor per iteration",
                     "Minimum lambda",
                     "Initial Gamma (smoothing)",
                     "Gamma factor per iteration",
                     "Minimum Gamma",
                     "Use positivity constraint"],
                    ["l", "c", "c", "e", "e", "e", "e", "e", "e", "e", "e",
                     "e", "c"],
                    [None, 1, 0, self.max_iter, self.max_diff_fac,
                     self.max_rel_diff, self.lam, self.lam_fac, self.lam_min,
                     self.gam, self.gam_fac, self.gam_min, None],
                    "Magnetic inversion parameters")
        else:
            results, okButton = dialog(
                    ["Maximum number of iterations",
                     "Maximum relative RMS misfit",
                     "Maximum variation of RMS",
                     "Initial lambda (regularization)",
                     "Lambda factor per iteration",
                     "Minimum lambda",
                     "Initial Gamma (smoothing)",
                     "Gamma factor per iteration",
                     "Minimum Gamma",
                     "Use positivity constraint"],
                    ["e", "e", "e", "e", "e", "e", "e", "e", "e", "c"],
                    [self.max_iter, self.max_diff_fac, self.max_rel_diff,
                     self.lam, self.lam_fac, self.lam_min,
                     self.gam, self.gam_fac, self.gam_min, None],
                    "Gravity inversion parameters")
        if not okButton:
            print("No inversion parameters given")
            return False
# set flags for properties to be optimized
        if "m" in data_type:
            if int(results[1]) > -1:
                self.sus_inv = True
            if int(results[2]) > -1:
                self.rem_inv = True
            if self.sus_inv and self.rem_inv:
                print("\nInverting for both, susceptibility and remanence, "
                      + "does not make sense.\n\nSusceptibility inversion "
                      + "canceled and susceptibility set to zero.")
                self.sus_inv = False
            if not self.sus_inv and not self.rem_inv:
                print("\nNo parameter type for inversion chosen.\n"
                      + "Susceptibility will be used by default")
                self.sus_inv = True
            ianswer = 3
        else:
            self.rho_inv = True
            ianswer = 0
# Set maximum number of iterations
        self.max_iter = int(results[ianswer])
        ianswer += 1
# set relative RMS misfit for all data sets which stops iterations if reached
        self.max_diff_fac = float(results[ianswer])
        ianswer += 1
# set maximum relative change of RMS misfit from one iteration to the next for
#     which iterations continue
        self.max_rel_diff = float(results[ianswer])
        ianswer += 1
# set relative importance of initial model and reduction of this parameter per
# iteration
        self.lam = float(results[ianswer])
        ianswer += 1
        self.lam_fac = float(results[ianswer])
        ianswer += 1
        self.lam_min = float(results[ianswer])
        ianswer += 1
        self.gam = float(results[ianswer])
        ianswer += 1
        self.gam_fac = float(results[ianswer])
        ianswer += 1
        self.gam_min = float(results[ianswer])
        ianswer += 1
        if int(results[ianswer]) > -1:
            self.positive = True
        else:
            self.positive = False
        return True

    def get_area2D(self):
        """
        Define model space and initial prism sizes for 2.5D model.
        If block sizes are reduced in certain areas, minimum accepted block
        sizes are stored in min_size_x, .._y, .._z

        Defines the following variables
        -------------------------------
        xprism_min: float
            Smallest x-coordinate for prism limits
        xprism_max: float
            Largest x-coordinate for prism limits
        dx_prism: float
            Initial prism size in x
        min_size_x: float
            Minimum allowed prism size in x

        Similar variable for the y and z direction
        """
    #    Calculate default prism sizes
        xmin = self.x.min()
        xmax = self.x.max()
        line_length = xmax - xmin
        dx = line_length/(len(self.x)-1)
        nr = 1-int(np.log10(line_length))
        while True:
            dx_ini = np.round(line_length/5., nr)
            if dx_ini > 0.:
                break
            nr += 1
        dz_ini = dx_ini/2.
        zmin = 0
        zmax = dz_ini
        yw = zmax

        nx = int(np.ceil(line_length/dx_ini))
        if dx_ini*nx-dx <= line_length:
            nx += 1
        xmin_p = xmin - zmax
        xmax_p = xmax + zmax

        results, okButton = dialog(
                ["xmin [m]", "xmax [m]", "dx_ini [m]", "minimum size x [m]",
                 "zmin [m]", "zmax [m]", "dz_ini [m]", "minimum size z [m]",
                 "prism_width [m]",
                 "Data point reduction\n(1 point out of...)"],
                ["e", "e", "e", "e", "e", "e", "e", "e", "e", "e"],
                [xmin_p, xmax_p, dx_ini, dx_ini/4., zmin, zmax, dz_ini,
                 dz_ini/4., yw, 1], "2D prism parameters")
        if not okButton:
            print("\nNo prism data given, inversion aborted")
            return False
        self.xprism_min = float(results[0])
        self.xprism_max = float(results[1])
        self.dx_prism = float(results[2])
        self.min_size_x = float(results[3])
        self.zprism_min = float(results[4])
        self.zprism_max = float(results[5])
        self.dz_prism = float(results[6])
        self.min_size_z = float(results[7])
        self.data_reduction = int(results[9])
        yw = float(results[8])
        self.yprism_min = -yw
        self.yprism_max = yw
        self.dy_prism = 2.*yw
        self.min_size_y = self.dy_prism
        self.prepare_data()
        self.set_prisms()
        return True

    def get_area3D(self):
        """
        Define model space and initial prism sizes for 3D model.
        If block sizes are reduced in certain areas, minimum accepted block
        sizes are stored in min_size_x, .._y, .._z

        Defines the following variables
        -------------------------------
        xprism_min: float
            Smallest x-coordinate for prism limits
        xprism_max: float
            Largest x-coordinate for prism limits
        dx_prism: float
            Initial prism size in x
        min_size_x: float
            Minimum allowed prism size in x

        Similar variable for the y and z direction

        """
#    Calculate default prism sizes
        xmin = self.x.min()
        xmax = self.x.max()
        line_length_x = xmax - xmin
        dx = line_length_x/(len(self.x)-1)
        ymin = self.y.min()
        ymax = self.y.max()
        line_length_y = ymax - ymin
        dy = line_length_y/(len(self.y)-1)
        length_xy = max((line_length_x, line_length_y))
        nr = 1-int(np.log10(length_xy))
        dx_ini = np.round(np.sqrt(line_length_x*line_length_y/25.), nr)
        dy_ini = dx_ini
        dz_ini = length_xy/10.
        zmin = 0.
        zmax = dz_ini
        nx = int(np.ceil(line_length_x/dx_ini))
        if dx_ini*nx-dx <= line_length_x:
            nx += 1
        xmin_p = xmin - dx_ini
        xmax_p = xmax + dx_ini

        ny = int(np.ceil(dy/dy_ini))
        if dy_ini*ny-dy <= line_length_y:
            ny += 1
        ymin_p = ymin - dy_ini
        ymax_p = ymax + dy_ini

# next lines are for testing purpose only
        # xmin_p = 680000.
        # xmax_p = 740000.
        # dx_ini = 15000.
        # ymin_p = 2240000.
        # ymax_p = 2285000.
        # dy_ini = 15000.
        # zmax = 5000.
        # dz_ini = 2500.

# Get initial prism configuration and prism control parameters
        results, okButton = dialog(
                ["xmin [m]", "xmax [m]", "dx_ini [m]", "minimum size x [m]",
                 "ymin [m]", "ymax [m]", "dy_ini [m]", "minimum size y [m]",
                 "zmin [m]", "zmax [m]", "dz_ini [m]", "minimum size z [m]",
                 "Data point reduction\n(1 point out of...)"],
                ["e", "e", "e", "e", "e", "e", "e", "e", "e", "e", "e", "e",
                 "e"],
                [xmin_p, xmax_p, dx_ini, dx_ini/4, ymin_p, ymax_p, dy_ini,
                 dy_ini/4, zmin, zmax, dz_ini, dz_ini/4, 1],
                "Prism parameters")
        if not okButton:
            print("\nNo prism data given, inversion aborted")
            return False
        self.xprism_min = float(results[0])
        self.xprism_max = float(results[1])
        self.dx_prism = float(results[2])
        self.min_size_x = float(results[3])
        self.yprism_min = float(results[4])
        self.yprism_max = float(results[5])
        self.dy_prism = float(results[6])
        self.min_size_y = float(results[7])
        self.zprism_min = float(results[8])
        self.zprism_max = float(results[9])
        self.dz_prism = float(results[10])
        self.min_size_z = float(results[11])
        self.data_reduction = int(results[12])
# Extract data within prism area and, if asked for, reduce point density
        self.prepare_data()
# Back up original (reduced) data
        self.data_ori = np.copy(self.data)
        self.std_data_ori = np.std(self.data)
        self.xmin = np.nanmin((self.x))
        self.xmax = np.nanmax((self.x))
        self.ymin = np.nanmin((self.y))
        self.ymax = np.nanmax((self.y))
        self.dx = (xmax-xmin)/(self.data_shape[1]-1)
        self.dy = (ymax-ymin)/(self.data_shape[0]-1)
        self.set_prisms()
        return True

    def prepare_data(self):
        """
        Extract data used for inversion from full data set and concatenate
        all data into one flattened numpy 1D array.
        Data reduction may be due to restriction of coordinates and/or due to
        reduction of number of data (taking only one point out of several in
        both directions).

        Reduction is done in situ, i.e. self.x, self.y, self.z, self.data and
        self.data_shape are modified.
        """
        if self.dim == 2 or len(self.data_shape) == 1:
            ndata = self.data1_shape[0]
            xdata_min = self.xprism_min + self.zprism_max
            xdata_max = self.xprism_max - self.zprism_max
            n1 = np.where(self.x[:ndata] >= xdata_min)[0][0]
            n2 = np.where(self.x[:ndata] <= xdata_max)[0][-1]+1
            n1 = max(n1, int(self.data_reduction/2))
            x = self.x[n1:n2:self.data_reduction]
            y = self.y[n1:n2:self.data_reduction]
            z = self.z[n1:n2:self.data_reduction]
            data = self.data[n1:n2:self.data_reduction]
            self.data_shape = data.shape
            self.data1_shape = data.shape
            self.n_data1 = len(data)
            if self.n_sensor == 2:
                n1 = np.where(self.x2[:ndata] >= xdata_min)[0][0]
                n2 = np.where(self.x2[:ndata] <= xdata_max)[0][-1]+1
                n1 = max(n1, int(self.data_reduction/2))
                x = np.concatenate((x, self.x2[n1:n2:self.data_reduction]))
                y = np.concatenate((y, self.y2[n1:n2:self.data_reduction]))
                z = np.concatenate((z, self.z2[n1:n2:self.data_reduction]))
                d2 = self.data2[n1:n2:self.data_reduction]
                self.n_data2 = len(d2)
                self.data2_shape = d2.shape
                data = np.concatenate((data, d2))
            else:
                self.n_data2 = 0
        else:
            ndata_y, ndata_x = self.data_shape
            ndata = ndata_x*ndata_y
            xdata_min = self.xprism_min + self.zprism_max
            xdata_max = self.xprism_max - self.zprism_max
            ydata_min = self.yprism_min + self.zprism_max
            ydata_max = self.yprism_max - self.zprism_max
            d = self.data.reshape(self.data_shape)
            x = self.x.reshape(self.data_shape)
            y = self.y.reshape(self.data_shape)
            z = self.z.reshape(self.data_shape)
            xcol = np.unique(self.x)
            yrow = np.unique(self.y)
            nx1 = np.where(xcol >= xdata_min)[0][0]
            nx2 = np.where(xcol <= xdata_max)[0][-1]+1
            nx1 = max(nx1, int(self.data_reduction/2))
            ny1 = np.where(yrow >= ydata_min)[0][0]
            ny2 = np.where(yrow <= ydata_max)[0][-1]+1
            ny1 = max(ny1, int(self.data_reduction/2))
            x = x[ny1:ny2:self.data_reduction, nx1:nx2:self.data_reduction].\
                flatten()
            y = y[ny1:ny2:self.data_reduction, nx1:nx2:self.data_reduction].\
                flatten()

            z = z[ny1:ny2:self.data_reduction, nx1:nx2:self.data_reduction].\
                flatten()

            data = d[ny1:ny2:self.data_reduction, nx1:nx2:self.data_reduction]
            self.data_shape = data.shape
            self.data1_shape = data.shape
            data = data.flatten()
            self.n_data1 = len(data)
            if self.n_sensor == 2:
                d = self.data2.reshape(self.data2_shape)
                x2 = self.x2.reshape(self.data2_shape)
                y2 = self.y2.reshape(self.data2_shape)
                z2 = self.z2.reshape(self.data2_shape)
                xcol = np.unique(self.x2)
                yrow = np.unique(self.y2)
                nx1 = np.where(xcol >= xdata_min)[0][0]
                nx2 = np.where(xcol <= xdata_max)[0][-1]+1
                nx1 = max(nx1, int(self.data_reduction/2))
                ny1 = np.where(yrow >= ydata_min)[0][0]
                ny2 = np.where(yrow <= ydata_max)[0][-1]+1
                ny1 = max(ny1, int(self.data_reduction/2))
                x = np.concatenate(
                    (x, x2[ny1:ny2:self.data_reduction,
                           nx1:nx2:self.data_reduction].flatten()))
                y = np.concatenate(
                    (y, y2[ny1:ny2:self.data_reduction,
                           nx1:nx2:self.data_reduction].flatten()))
                z = np.concatenate(
                    (z, z2[ny1:ny2:self.data_reduction,
                           nx1:nx2:self.data_reduction].flatten()))
                d2 = d[ny1:ny2:self.data_reduction,
                       nx1:nx2:self.data_reduction]
                self.data2_shape = d2.shape
                data = np.concatenate((data, d2.flatten()))
                self.n_data2 = d2.shape[0] * d2.shape[1]
        self.x = x
        self.y = y
        self.z = z
        self.data = data
        print(f"\n{len(data)} data points to be inverted")
        self.data_ori = np.copy(data)
        self.data_ori -= np.nanmedian(self.data_ori)
        self.std_data_ori = np.std(self.data_ori)
        self.dx = self.x[1] - self.x[0]
        self.dy = self.y[1] - self.y[0]
        self.xplt_min = self.x.min()-self.dx*0.5
        self.xplt_max = self.x.max()+self.dx*0.5
        self.yplt_min = self.y.min()-self.dy*0.5
        self.yplt_max = self.y.max()+self.dy*0.5
        if max(self.xprism_max, self.yprism_max) > 100000:
            self.xplt_min /= 1000.
            self.xplt_max /= 1000.
            self.yplt_min /= 1000.
            self.yplt_max /= 1000.
            self.ax_unit = "km"
        else:
            self.ax_unit = "m"
        return True

    def set_plot_depths(self):
        """
        Calculate depths at which horizontal planes through the final model
        shall be plotted

        Returns
        -------
        None.

        """
# Define depths at which resulting model should be plotted (list of depths)
        self.z_plot = np.arange(self.zprism_min+self.min_size_z/2,
                                self.zprism_max, self.min_size_z)
        self.nz_plot = []
        for i, zz in enumerate(self.z_plot):
            self.nz_plot.append(int((zz-self.zprism_min)/self.min_size_z))

    def set_prisms(self):
        """
        Define prisms for initial model

        Returns
        -------
        None.

        """
# Define initial model
        self.x_prism = np.arange(self.xprism_min,
                                 self.xprism_max+self.dx_prism/2,
                                 self.dx_prism)
        self.z_prism = np.arange(self.zprism_min,
                                 self.zprism_max+self.dz_prism/2,
                                 self.dz_prism)
        self.mod_xshape = len(self.x_prism)-1
        self.mod_zshape = len(self.z_prism)-1
        if self.dim == 3:
            self.y_prism = np.arange(self.yprism_min,
                                     self.yprism_max+self.dy_prism/2,
                                     self.dy_prism)
            self.yprism_topo = np.copy(self.y_prism)
            self.mod_yshape = len(self.y_prism)-1
        else:
            self.y_prism = np.array([self.yprism_min, self.yprism_max])
            # self.y_prism_topo = self.y_prism+
            self.mod_yshape = 1
# Define prisms of initial model
        self.mPrism = PP(self.earth, self.min_size_x, self.min_size_y,
                         self.min_size_z)
        self.nx_prism = len(self.x_prism)-1
        self.ny_prism = len(self.y_prism)-1
        self.nz_prism = len(self.z_prism)-1
        self.n_prisms = self.nx_prism*self.ny_prism*self.nz_prism
        self.prism_nr = []
        if self.topo_flag:
            for i in range(self.mod_xshape):
                xpr = np.array([self.x_prism[i], self.x_prism[i+1]])
                xtopo = np.array([self.x_prism[i], self.x_prism[i+1],
                                  self.x_prism[i+1], self.x_prism[i]])
                for j in range(self.mod_yshape):
                    ypr = np.array([self.y_prism[j], self.y_prism[j+1]])
                    ytopo = np.array([self.y_prism[j], self.y_prism[j],
                                      self.y_prism[j+1],  self.y_prism[j+1]])
                    ztopo = self.topo_int(xtopo, ytopo)
                    ztmin = ztopo.max()
                    z1 = self.zprism_max
                    while True:
                        z2 = z1-self.dz_prism
                        if z2 < ztmin:
                            zpr = np.array([z2, z1])
                        else:
                            zpr = np.array([list(ztopo) + [z1]])
                        z1 -= self.dz_prism
        else:
            for i in range(self.mod_xshape):
                xpr = np.array([self.x_prism[i], self.x_prism[i+1]])
                for j in range(self.mod_yshape):
                    ypr = np.array([self.y_prism[j], self.y_prism[j+1]])
                    for k in range(self.mod_zshape):
                        zpr = np.array([self.z_prism[k], self.z_prism[k+1]])
                        self.mPrism.add_prism(xpr, ypr, zpr, 0.001, 0.1,
                                              self.earth.inc, self.earth.dec,
                                              10.)
        print("Model prism dictionary defined with "
              + f"{len(self.mPrism.prisms)} prisms")
# Prepare book-keeping arrays
        if self.sus_inv and self.rem_inv:
            self.par_hist.append((np.zeros(self.mPrism.n_prisms), 2))
        else:
            self.par_hist.append((np.zeros(self.mPrism.n_prisms), 1))
        self.param_prism += list(range(self.mPrism.n_prisms))
        self.i0 = self.mPrism.n_prisms
        self.n_param = self.mPrism.n_prisms
        self.params = np.zeros(self.n_param+1)
        i = -1
        if self.sus_inv:
            for key, val in self.mPrism.prisms.items():
                i += 1
                self.params[i] = val.sus
        if self.rem_inv:
            for key, val in self.mPrism.prisms.items():
                i += 1
                self.params[i] = val.rem
        if self.rho_inv:
            for key, val in self.mPrism.prisms.items():
                i += 1
                self.params[i] = val.rho

        return True

    def get_variances(self):
        """
        Get variances for data and parameters for normalization of inversion
        matrices

        Defines the following variables
        -------------------------------
        sigma_mag, sigma_grav : floats
            Uncertainty of magnetic data [nT] and gravity data [mGal])
        sigma_sus, sigma_rem, sigma_rho : floats
            Variability of susceptibilities [SI], remanence [A/m],
            densities [kg/m3]
        depth_ref : float
            Reference depth for decrease of sensitivity of prism parameters.
            For sus & rem, sensitivity (regularization) decreases with
            (depth_ref/depth)**3.
            For density, sensitivity (regularization) decreases with
            (depth_ref/depth)**2.
        width_max : int
            Number of samples around every data point (to each side) defining
            the area over which the sample value must be larger than the
            neighbours to be considered as local maximum
        max_amp : float
            maximum relative amplitude [%/100] to be considered for a local
            anomaly in order to split prims (if max_val[i]/(max(max_val
            min(max_val)) > max_amp, prisms are split into up to 8 smaller
            prisms (division by 2 in all directions) for the next iteration)

        """
        if self.sus_inv:
            if self.rem_inv:
                results, okButton = dialog(
                        ["sigma_mag",
                         "\n",
                         "sigma_sus",
                         "sigma_rem",
                         "\n",
                         "sigma factor at bottom\n   will be cubed",
                         "width_max [samples]",
                         "max_amp [%]"],
                        ["e", "l", "e", "e", "l", "e", "e", "e"],
                        [self.sigma_mag, None, self.sigma_sus, self.sigma_rem,
                         None, 2., self.width_max,
                         self.max_amp*100.], "Inversion parameters")
                if not okButton:
                    print("\nNo inversion done: no inversion parameters given")
                    return False
                self.sigma_mag = float(results[0])
                self.sigma_sus = float(results[2])
                self.sigma_rem = float(results[3])
                self.depth_ref = float(results[5])
                self.width_max = int(results[6])
                self.max_amp = float(results[7])/100.
            else:
                results, okButton = dialog(
                        ["sigma_mag",
                         "\n",
                         "sigma_sus",
                         "\n",
                         "sigma factor at bottom\n   will be cubed",
                         "width_max [samples]",
                         "max_amp [%]"],
                        ["e", "l", "e", "l", "e", "e", "e"],
                        [self.sigma_mag, None, self.sigma_sus, None,
                         2., self.width_max, self.max_amp*100.],
                        "Inversion parameters")
                if not okButton:
                    print("\nNo inversion done: no inversion parameters given")
                    return False
                self.sigma_mag = float(results[0])
                self.sigma_sus = float(results[2])
                self.depth_ref = float(results[4])
                self.width_max = int(results[5])
                self.max_amp = float(results[6])/100.
        elif self.rem_inv:
            results, okButton = dialog(
                    ["sigma_mag",
                     "\n",
                     "sigma_rem",
                     "\n",
                     "sigma factor at bottom\n   will be squared",
                     "width_max [samples]",
                     "max_amp [%]"],
                    ["e", "l", "e", "l", "e", "e", "e"],
                    [self.sigma_mag, None, self.sigma_rem, None,
                     self.dz_prism/10., self.width_max, self.max_amp*100.],
                    "Inversion parameters")
            if not okButton:
                print("\nNo inversion done: no inversion parameters given")
                return False
            self.sigma_mag = float(results[0])
            self.sigma_rem = float(results[2])
            self.depth_ref = float(results[4])
            self.width_max = int(results[5])
            self.max_amp = float(results[6])/100.
        else:
            results, okButton = dialog(
                    ["sigma_grav",
                     "\n",
                     "sigma_rho",
                     "\n",
                     "depth_ref [m]",
                     "width_max [samples]",
                     "max_amp [%]"],
                    ["e", "l", "e", "l", "e", "e", "e"],
                    [self.sigma_grav, None, self.sigma_rho, None,
                     self.dz_prism/10., self.width_max, self.max_amp*100.],
                    "Inversion parameters")
            if not okButton:
                print("\nNo inversion done: no inversion parameters given")
                return False
            self.sigma_grav = float(results[0])
            self.sigma_rho = float(results[2])
            self.depth_ref = float(results[4])
            self.width_max = int(results[5])
            self.max_amp = float(results[6])/100.
        return True

    def write_parameters(self, file):
        """
        Save inversion control parameters to file parameters_date&time.dat

        Parameters
        ----------
        file : str
            name of output file

        Returns
        -------
        None.

        """
        with open(file, "w") as fo:
            fo.write("File(s):\n")
            # for i,fi in enumerate(data_files):
            #     fo.write(f"   {fi} as type {type_values[data_types[i]]}, "
            #              +f"measurement height {sensor_heights[i]}\n")
            fo.write(f"\nEarth's field: strength: {self.earth.f} nT, "
                     + f"inclination: {self.earth.inc} deg, "
                     + f"declination: {self.earth.dec} deg\n")
            fo.write("\nParameters\n")
            if self.sus_inv:
                fo.write("   Invert for susceptibilities\n")
            if self.rem_inv:
                fo.write("   Invert for remanence\n")
            if self.rho_inv:
                fo.write("   Invert for densities\n")
            fo.write(f"   Maximum number of iterations: {self.max_iter}\n")
            fo.write(f"   Stop if relative misfit < {self.max_diff_fac*100}% "
                     + "or relative misfit variation < "
                     + f"{self.max_rel_diff*100}%\n")
            fo.write(f"   Lambda: Initial: {self.lam}, minimum: " +
                     f"{self.lam_min}, factor per iteration: {self.lam_fac}\n")
            fo.write(f"   Gamma:  Initial: {self.gam}, minimum: {self.gam_min}"
                     + f", factor per iteration: {self.gam_fac}\n")
            fo.write("\nInitial prism area\n")
            fo.write(f"   West:  {self.xprism_min}, East:  {self.xprism_max}, "
                     + f"initial size in X: {self.dx_prism}, "
                     + f"min_size_x: {self.min_size_x}\n")
            fo.write(f"   South: {self.yprism_min}, North: {self.yprism_max}, "
                     + f"initial size in Y: {self.dy_prism}, "
                     + f"min_size_y: {self.min_size_y}\n")
            fo.write(f"   Top:   {self.zprism_min}, Bottom: {self.zprism_max},"
                     + f" initial size in Z: {self.dz_prism}, "
                     + f"min_size_z: {self.min_size_z}\n")
            fo.write("\nData uncertainty:\n")
            if "m" in self.data_type:
                fo.write(f"   Magnetics: {self.sigma_mag}\n")
            if self.data_type == "g":
                fo.write(f"   Gravity:   {self.sigma_grav}\n")
            fo.write("\nParameter variability:\n")
            if self.sus_inv:
                fo.write(f"   Susceptibility: {self.sigma_sus}\n")
            if self.rem_inv:
                fo.write(f"   Remancence: {self.sigma_rem}\n")
            if self.rho_inv:
                fo.write(f"   Gravity:   {self.sigma_rho}\n")
            fo.write("   Reference depth for weighting of parameter "
                     + f"variability: {self.depth_ref}\n")
            fo.write("   Size of area around a point for local maximum "
                     + f"determination: {self.width_max} points\n")
            fo.write("   Minimum relative amplitude for prism split: "
                     + f"{self.max_amp*100}%\n")



# -*- coding: utf-8 -*-
"""
Last mofified on Nov 22, 2024

@author: Hermann Zeyen <hermann.zeyen@gmail.com>
         Université Paris-Saclay
"""

from PyQt5 import QtWidgets, QtCore
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QRadioButton, QButtonGroup
import numpy as np


def dialog(labels, types, values, title="Title"):
    """
    Wrapper for class Dialog. Two buttons are shown for finishing: Ok and
    Cancel.

    Parameters
    ----------
    labels : list of strings
        Explanatory text put beside the fields for data entrance.
        If values==None, label is interpreted as information/Warning text.
        For a series of Radiobuttons or Comboboxes, all corresponding labels
        are given as a list within the list.
    types : list of strings (length as labels).
        Possible values:

        - "c": to define a check box;
        - "e": to define an editable box (lineEdit);
        - "l": To define a simple label, comment text;
        - "r": to define a series of radiobuttons (the corresponding list of
          labels is considered as one single label for the numbering of
          labels, types and values);
        - "b": to define a combobox (dropdown menu).

    values : list of texts, numbers or Nones (length as labels)
        Initial values given to editable fields.

        - Initial values for LineEdit fields (float, int or str)
        - None for check boxes and comboboxes
        - Number of radiobutton to be activated by default (natural numbering,
          starting at 1, not at 0).
        - For labels, it may be "b" (bold text), "i" (italic) or anything else,
          including None for standard text.

    title : str, default: "Title"
        Title of the dialog box.

    Returns
    -------
    results : list of str
        Response of each data entrance field. Should be transformed to the
        needed data format (int, float, ...) in the calling function.

        - For radiobuttons, the returned value indicates the number of the
          active button (counting starts at zero).
        - For checkboxes, the returned value is -1 if the box was not checked
          or gives the order in which the box was checked (starting at zero for
          the first box having been checked).

    Dbutton: bool
        If True, "Apply" button has been pressed to finish dialog, if False
        "Cancel" button has been pressed.

    """
    D = Dialog(labels, types, values, title)
    D.Dfinish = False
    D.exec_()
    while not D.Dfinish:
        QtCore.QCoreApplication.processEvents()

    results = []
    iline = 0
    n = len(values)
    if n > 0:
        for it, t in enumerate(types):
            if t.lower() == "e":
                results.append(D.dlines[iline].text())
                iline += 1
            elif t.lower() == "r":
                results.append(None)
                for i in range(len(labels[it])):
                    iline += 1
                    if D.rbtn[it][i].isChecked():
                        results[-1] = i
            #                            break
            elif t.lower() == "c":
                results.append(D.ck_order[it] - 1)
                iline += 1
            elif t.lower() == "b":
                results.append(D.combo[it].currentIndex())
                iline += 1
            else:
                results.append(None)
                iline += 1
    return results, D.Dbutton


class Dialog(QtWidgets.QDialog):
    """
    Created on Tue Sep 13 12:50:26 2022
    @author: Hermann Zeyen

    General purpose dialog box

    Input
    -----
    labels (list, string)
        labels explaining the different fields. For a series of radiobuttons or
        combobax, labels[i] is itself a list.
    types (list,string)
        In the order of "labels", indicates the type of field the label belongs
        to. If labels[i] is itself a list, it is considered as one single label

        String may be:

        - 'l' for label (no dialog entry, only information)
        - 'c' for checkbox
        - 'e' for LineEdit (to enter values)
        - 'r' for radiobutton
        - 'b': to define a combobox (dropdown menu).

        May be capital or not
    values (list, string, float or int); Optional, default value: None
        - Initial values for LineEdit fields;
        - For Radiobuttons: number of button activated by default (natural
          counting);
        - for labels, it may be "b" (bold), "i" (italic), or anything else
          (usually None) for standard text.
        - Ignored for checkbox and combobox.

    title (string). Optional, default value: "Title"
        Title of the dialog window.

    """

    def __init__(self, labels, types, values=None, title="Title"):
        super().__init__()
        nlab = len(labels)
        self.labels = labels
        self.ck_order = np.zeros(nlab, dtype=int)
        self.n_checked = 0
        self.Dfinish = False
        self.Dbutton = False
        self.dlabels = []
        self.dlines = []
        self.ckState = []
        self.ckb = []
        self.rbtn = []
        self.btngroup = []
        self.combo = []
        for i in range(nlab):
            self.ckState.append(False)
        self.YesBtn = QtWidgets.QPushButton("Ok", self)
        self.YesBtn.move(10, 20 * (nlab + 3))
        self.CancelBtn = QtWidgets.QPushButton("Cancel", self)
        self.CancelBtn.move(150, 20 * (nlab + 3))
        self.mainLayout = QtWidgets.QGridLayout()
        self.setLayout(self.mainLayout)
        il_add = 0
        ilin = 0
        for i, lab in enumerate(labels):
            il = ilin + il_add
            if types[i].lower() == "l":
                if values[i]:
                    if values[i].lower() == "b":
                        self.dlabels.append(QtWidgets.QLabel("<b>" + lab +
                                                             "</b>"))
                    elif values[i].lower() == "i":
                        self.dlabels.append(QtWidgets.QLabel("<i>" + lab +
                                                             "</i>"))
                    else:
                        self.dlabels.append(QtWidgets.QLabel(lab))
                else:
                    self.dlabels.append(QtWidgets.QLabel(lab))
                self.dlines.append(None)
                self.ckb.append(None)
                self.rbtn.append(None)
                self.btngroup.append(None)
                self.combo.append(None)
                self.mainLayout.addWidget(self.dlabels[ilin], il, 0, 1, 2)
                ilin += 1
            elif types[i].lower() == "e":
                self.dlabels.append(QtWidgets.QLabel(lab))
                self.dlines.append(QtWidgets.QLineEdit())
                self.ckb.append(None)
                self.rbtn.append(None)
                self.btngroup.append(None)
                self.combo.append(None)
                self.mainLayout.addWidget(self.dlabels[ilin], il, 0, 1, 1)
                self.mainLayout.addWidget(self.dlines[ilin], il, 1, 1, 1)
                # try:
                s = str(values[i])
                self.dlines[-1].setText(s)
                # except:
                #     pass
                ilin += 1
            elif types[i].lower() == "r":
                self.ckb.append(None)
                self.combo.append(None)
                self.rbtn.append([])
                self.btngroup.append(QButtonGroup())
                rck = int(values[i]) - 1
                if rck < 0 or rck >= len(lab):
                    rck = 0
                for ir, la in enumerate(lab):
                    self.dlabels.append(None)
                    self.dlines.append(None)
                    self.rbtn[i].append(QRadioButton(la))
                    self.btngroup[-1].addButton(self.rbtn[i][-1])
                    self.mainLayout.addWidget(self.rbtn[i][-1], il, 0, 1, 2)
                    if ir == rck:
                        self.rbtn[i][-1].setChecked(True)
                    else:
                        self.rbtn[i][-1].setChecked(False)
                    il += 1
                    ilin += 1
            elif types[i].lower() == "c":
                self.dlabels.append(None)
                self.dlines.append(None)
                self.rbtn.append(None)
                self.btngroup.append(None)
                self.combo.append(None)
                self.ckb.append(QtWidgets.QCheckBox(self))
                self.ckb[i].setText(lab)
                self.mainLayout.addWidget(self.ckb[i], il, 0, 1, 2)
                self.ckb[i].stateChanged.connect(self.checked)
                if values[i]:
                    self.ckb[i].setChecked(True)
                    self.ckState[i] = True
                    self.ck_order[i] = max(self.ck_order) + 1
                ilin += 1
            elif types[i].lower() == "b":
                self.dlabels.append(None)
                self.dlines.append(None)
                self.ckb.append(None)
                self.rbtn.append(None)
                self.btngroup.append(None)
                self.combo.append(QtWidgets.QComboBox())
                for il, la in enumerate(lab):
                    self.combo[i].addItem(la)
                ilin += 1
                self.mainLayout.addWidget(self.combo[i], ilin, 0, 1, 1)
                il_add += 1
        ilin += 2
        il = ilin + il_add
        self.mainLayout.addWidget(self.YesBtn, il, 0)
        self.mainLayout.addWidget(self.CancelBtn, il, 1)
        self.YesBtn.setDefault(True)
        self.YesBtn.clicked.connect(self.on_YesButton_clicked)
        self.CancelBtn.clicked.connect(self.on_CancelButton_clicked)

        self.setWindowTitle(title)
        self.setWindowFlags(QtCore.Qt.WindowStaysOnTopHint)
        self.setModal(True)
        self.show()

    def checked(self, checked):
        """
        Actions executed if a check box has changed its state.
        If a box has been checked, the function searches the one which was
        checked using self.ckState as indicator (this variable contains the
        state of all check boxes before the click). Its click-order is stored
        and self.ckState is changed. In addition, if the calling function is
        "False_Color", the explanation text is changed and the color to be
        used for the following clicked item is indicated. If in this case
        the third box is clicked, a message appears that no other box can be
        checked unless one is unchecked. If nevertheless a forth box is
        checked, the corresponding box is automatically unchecked.
        If a box is unchecked, this is stored in self.ckState and the colors
        indicated are changed if necessary and if function is "False_Color".

        Parameters
        ----------
        checked : QtWidget.QCheckBox.checkState
            state of a checkbox after clicking into it

        Returns
        -------
        None.

        """
# If a nex box is checked, search the one which has been checked
# If self.ckb has None value, the corresponding entry is not a checkbox
        if checked == Qt.Checked:
            for i, ck in enumerate(self.ckb):
                if not ck:
                    continue
# If self.ckb.checkState is checked after click, set ckState to True and do
#    changes
                if ck.checkState() == Qt.Checked:
                    self.ckState[i] = True
# If checkbox nr i was not checked, increase the number of checked boxes
#    (n_checked) and store the order of checkin in self.ck_order
                    if self.ck_order[i] == 0:
                        self.n_checked += 1
                        self.ck_order[i] = self.n_checked
# if checkboxes are called from function "False_Color", change the label,
#    indicating the clor which will be used for the corresponding indicator.
                        self.ckb[i].setText(
                            f"{self.labels[i]} (" + f"{self.ck_order[i]})")
                        break
# If self.ckb.checkState is still unchecked, set ckState to Falsee
                else:
                    self.ckState[i] = False
# If click has unchecked a checkbox, do necessary changes
# If self.ckb has None value, the corresponding entry is not a checkbox
        else:
            for i, ck in enumerate(self.ckb):
                if not ck:
                    continue
# If self.ckb.checkState is still checked, set ckState to True
                if ck.checkState() == Qt.Checked:
                    self.ckState[i] = True
# If checkbox is no longer checked but it was (self?ckState), the unchecked box
#    is found
                else:
                    if self.ckState[i] is True:
                        self.ckState[i] = False
                        n = self.ck_order[i]
# reset ck_order to 0 (indicating also unchecked box)
                        self.ck_order[i] = 0
# Reset label to initial value (changes only for function "False_Color")
                        self.ckb[i].setText(self.labels[i])
# For all boxes that were checked later than the unchecked one, reduce their
#    checking order by 1 and, if function is False_Color, indicate the new
#    color uses for plotting
                        for j, ck in enumerate(self.ck_order):
                            if ck > n:
                                ck -= 1
                                self.ckb[j].setText(f"{self.labels[j]} (" +
                                                    f"{ck})")
                        self.n_checked -= 1
                        break
        self.show()

    def on_YesButton_clicked(self):
        """
        If Ok button has been clicked, close Dialogue and set Dbutton to True.
        """
        n_checked = 0
        for ck in self.ckb:
            if not ck:
                continue
            if ck.checkState() == Qt.Checked:
                n_checked += 1
        self.Dfinish = True
        self.Dbutton = True
        self.close()

    def on_CancelButton_clicked(self):
        """
        If Cancel button has been clicked, close Dialogue and set Dbutton to
        False.
        """
        self.Dfinish = True
        self.Dbutton = False
        self.close()



# -*- coding: utf-8 -*-
"""
Created on Tue Nov 19 18:07:18 2024

@author: Hermann
"""
import numpy as np
from PyQt5 import QtCore, QtWidgets
from PyQt5.QtWidgets import QVBoxLayout, QWidget
import matplotlib.pyplot as plt
from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg as FigureCanvas,
    NavigationToolbar2QT as NavigationToolbar)
from matplotlib.gridspec import GridSpec


class newWindow(QWidget):
    """
    This "window" is a QWidget. If it has no parent, it will appear as a
    free-floating window.
    
    contains the following methods:
        - __init__
        - on_enter_event : give focus to window if mouse enters it or moves
            inside
        - setHelp : Write help text at bottom of screen
        - get_event : get keyboard or mouse event
        - follow_line : Follow cursor and plots a line from former point
        - plot_image : Plot a matplotlib imshow image into a subplot
        - plot_images : Plots up to 3 images into 3 subplots
    """

    def __init__(self, title, xsize=1800, ysize=1200, width=20, height=12):
        """
        Initialize a floating window

        Parameters
        ----------
        title : str
            Title of figure window
        xsize, ysize : integers, optional. Defaults: 1800 and 1200
            width and height of figure window in pixels
        width, height : floats, optional. Defaults 15 and 9
            width and height of figure itself in cm
        """
        super().__init__()
        self.fig = plt.figure(figsize=(width, height))
        self.fig.tight_layout()
        self.layout = QVBoxLayout()
        self.setLayout(self.layout)
        self.resize(xsize, ysize)
        self.setWindowTitle(title)
        self.canvas = FigureCanvas(self.fig)
        self.layout.addWidget(self.canvas)
        self.canvas.setFocusPolicy(QtCore.Qt.ClickFocus)
        self.cid_enter = self.canvas.mpl_connect('axes_enter_event',
                                                 self.on_enter_event)
        self.canvas.draw()
        self.toolbar = NavigationToolbar(self.canvas, self)
        self.layout.addWidget(self.toolbar)
        self.xsize = xsize
        self.ysize = ysize
        self.width = width
        self.height = height

    def on_enter_event(self, event):
        self.canvas.setFocus()

    def setHelp(self, text):
        """
        Set help text at bottom of screen.

        Parameters
        ----------
        text : str
            Text to be printed (defined in __init__)
            Text is written in QLabel widget. In order to keep the widget at
            the bottom of the screen, the existing one is first erased and
            then reopened. This call has to be done after any modification of
            the graphics window.

        fig : Matplotlib Figure
            Figure where to write the help text.

        Returns
        -------
        None
        """
        try:
            self.help.close()
        except (AttributeError, RuntimeError):
            pass
        self.help = QtWidgets.QLabel(self)
        self.help.setMaximumHeight(20)
        self.layout.addWidget(self.help)
        self.help.show()
        self.help.setText(text)

    def get_event(self):
        """
        wait for a mouse click of keyboard press within figure "fig".

        Parameters
        ----------
        fig : matplotlib figure
            for main window, it must be self.w.mplvl
            for floating window, it is the name that has been given to the
            window at creation time

        Returns
        -------
        event.
        Mouse position, button pressed and the general event parameters are
        accessible through event.xdata, event.ydata, event.button and
        others like event.x or event.y

        """
        self.wait = True

        def onClick(event):
            self.wait = False
            self.event = event

        self.canvas.mpl_connect("button_press_event", onClick)
        self.canvas.mpl_connect("key_press_event", onClick)
        while self.wait:
            QtCore.QCoreApplication.processEvents()
        return self.event

    def follow_line(self, ax, release_flag=False, nleft=1, nright=1):
        """
        Pull line across plot
        Parameters
        ----------
        fig, ax : names of figures and axis where to follow the line
        release_flag (boolean): If True, end of line when left button released\
                                if False, end of line triggered by pressing
                                right button
        nleft (int): if 0 start line for negative direction at origin \
                     if not, start line at the position of first click
        nright (int): if 0 start line for positive direction at origin \
                     if not, start line at the position of first click

        Returns
        -------
        event : QT mouse event
            Information concerning the last mouse event.
            Mainly useful are event.button (mouse button pressed/released);
            event.xdata and event.ydata: coordinates of last mouse position in
            axis coordinate system (e.g. meters);
            event.x and event.y: coordinates of last mouse position in
            window coordinates (pixels)

        coor_x and coor_y: list of floats
            coordinates (axis coordinate system) of line segment(s)

        Mouse button pressed to exit sunction (may be 2 for wheel ar 3
        for right button)

        Left button adds new point. right button finishes. Wheel erases last
        clicked point or, if no points are available, returns
        """
# set flags and initialize coordinates
        self.released = False
        self.start = []
        self.coor_x = []
        self.coor_y = []
        figure = self.fig
        (self.line,) = ax.plot(self.coor_x, self.coor_y, "k", animated=True)

        def onPress(event):
            self.event = event
            self.line_click = False
# left mouse button is pressed
            if event.button == 1:
                if event.xdata is None or event.ydata is None:
                    self.mouse = 1
                    self.x_event = event.xdata
                    self.y_event = event.ydata
                    return
                if len(self.coor_x) == 0:
                    if (event.xdata < 0 and nleft == 0) or (
                            event.xdata >= 0 and nright == 0):
                        self.start = [0, 0]
                        self.coor_x.append(0)
                        self.coor_y.append(0)
                    else:
                        self.start = [event.xdata, event.ydata]
                        self.coor_x.append(event.xdata)
                        self.coor_y.append(event.ydata)
                    if event.xdata < 0:
                        self.side = -1
                    else:
                        self.side = 1
                    self.background = self.fig.canvas.copy_from_bbox(
                        self.fig.bbox)
# set starting point initially also as end point
                self.coor_x.append(event.xdata)
                self.coor_y.append(event.ydata)
                self.canvas_follow = self.line.figure.canvas
                self.axl = self.line.axes
                self.line.set_data(self.coor_x, self.coor_y)
                self.axl.draw_artist(self.line)
# set action on mouse motion
                self.cidmotion = self.line.figure.canvas.mpl_connect(
                    "motion_notify_event", onMotion)
# set action on mouse release
                if release_flag:
                    self.cidrelease = self.line.figure.canvas.mpl_connect(
                        "button_release_event", onRelease)
# if right button is pressed, finish
            elif event.button == 3:
                self.mouse = 3
                self.x_event = event.xdata
                self.y_event = event.ydata
                try:
                    self.line.figure.canvas.mpl_disconnect(self.cidpress)
                except NameError:
                    pass
                self.line_click = False
                # try:
                #     self.line.figure.canvas.mpl_disconnect(self.cidmotion)
                # except NameError:
                #     pass
                if len(self.coor_x) > 0:
                    self.line_click = True
                self.line.set_animated(False)
                if self.line_click:
                    self.fig.canvas.restore_region(self.background)
                self.background = None
                self.released = True
                return
# Wheel is pressed, erase last point
            else:
                if len(self.coor_x) > 0:
                    print(f"Erase point ({self.coor_x[-1]},{self.coor_y[-1]})")
                    del self.coor_x[-1]
                    del self.coor_y[-1]
                    self.canvas_follow = self.line.figure.canvas
                    self.axl = self.line.axes
                    self.line.set_data(self.coor_x, self.coor_y)
                    self.axl.draw_artist(self.line)
# set action on mouse motion
                    self.cidmotion = self.line.figure.canvas.mpl_connect(
                        "motion_notify_event", onMotion)
                else:
                    self.mouse = 2
                    self.x_event = event.xdata
                    self.y_event = event.ydata
                    try:
                        self.line.figure.canvas.mpl_disconnect(self.cidpress)
                    except (NameError, AttributeError):
                        pass
                    self.line_click = False
                    try:
                        self.line.figure.canvas.mpl_disconnect(self.cidmotion)
                    except (NameError, AttributeError):
                        pass
                    if len(self.coor_x) > 0:
                        self.line_click = True
                    self.line.set_animated(False)
                    if self.line_click:
                        self.fig.canvas.restore_region(self.background)
                    self.background = None
                    self.released = True
                    return

        def onRelease(event):
            """
            When mouse is released, finish line segment

            Parameters
            ----------
            event : QT mouse event
                Not really needed, but necessary for routine call

            Returns
            -------
            None.

            """
# If line finishes when button is released do this here
            global figure
            self.line.figure.canvas.mpl_disconnect(self.cidpress)
            self.line.figure.canvas.mpl_disconnect(self.cidmotion)
            self.line.figure.canvas.mpl_disconnect(self.cidrelease)
            self.line.set_animated(False)
            if len(self.coor_x) > 0:
                self.fig.canvas.restore_region(self.background)
            self.background = None
            self.released = True
            return False

        def onMotion(event):
            """
            When mouse is moved, redraw the line from the fixed starting point
            to the new mouse position

            Parameters
            ----------
            event : QT mouse event
                contains mainly coordinates of new mouse position

            Returns
            -------
            None.

            """
            global figure
            self.event = event
            if event.xdata is None or event.ydata is None:
                return False
# set second point of line as actual mouse position
            self.coor_x[-1] = event.xdata
            self.coor_y[-1] = event.ydata
# Draw new line
            self.line.set_data(self.coor_x, self.coor_y)
            self.line.set_color("k")
            self.canvas_follow = self.line.figure.canvas
            self.axl = self.line.axes
            self.fig.canvas.restore_region(self.background)
            self.axl.draw_artist(self.line)
            self.canvas_follow.blit(self.axl.bbox)
            return True

# set action on mouse motion
        # self.cidmotion = self.line.figure.canvas.mpl_connect(
        #     "motion_notify_event", onMotion)
        self.cidpress = self.line.figure.canvas.mpl_connect(
            "button_press_event", onPress)
# As long as release flag is not set listen to events
        while self.released is not True:
            QtCore.QCoreApplication.processEvents()
# Return event information and vector of line coordinates
        return self.event, self.coor_x, self.coor_y

    def plot_image(self, ax, data, x, y, mincol, maxcol, percent, c, ptitle,
                   xlabel, ylabel, clabel, grid_flag, fontsize, bar_or,
                   nticks, bar_height):
        """
        Plot one 2D arrays into floating window using matplotlib's imshow.

        Parameters
        ----------
        ax : Matplotlib Axis
            Axis where to plot the image
        data : 2D numpy float array
        x : numpy 1D float array
            Positions of the columns
        y : numpy 1D float array
            Positions of the rows
        mincol : float
            Minimum value of color scale
        maxcol : float
            Maximum value of color scale.
            If mincol == maxcol,  and percent == 0, the limits of the color
            scale are minimum and maximum values of the data.
        percent : float
            If >0, mincol and maxcol are calculated as quantiles
            (percent, 1-percent).
        c: str
            color scale to be used.
        ptitle : str
            Plot title
        xlabel : str
            Lable of horizontal axis
        ylabel : str
            Lable of vertical axis
        clabel : str
            Lable of color bar
        grid_flag : bool
            If True, grid of major ticks is plotted
        fontsize : int
            Fontsize of title. Fontsizes of all other labels or numbers are
            reduced by 2 points
        bar_or : str
            Color bar orientation? May be "horizontal" of "vertical"
        nticks : int
            approximate number of ticks desired for color bar

        Returns
        -------
        ax_float: List of Matplot axes

        """
        rd = int(np.ceil(-np.log10(np.nanmax(abs(data)))))+2
        if mincol == maxcol:
            if percent > 0:
                max_col = np.round(np.nanquantile(data, 1-percent), rd)
                min_col = np.round(np.nanquantile(data, percent), rd)
            else:
                max_col = np.nanmax(data)
                min_col = np.nanmin(data)
        else:
            max_col = maxcol
            min_col = mincol
        dx2 = (x[1]-x[0])/2
        dy2 = (y[1]-y[0])/2
        im1 = ax.imshow(np.flip(data, axis=0), cmap=c, aspect="equal",
                        extent=[np.min(x)-dx2, np.max(x)+dx2, np.min(y)-dy2,
                                np.max(y)+dy2], vmin=min_col, vmax=max_col)
        ax.set_title(ptitle, fontsize=fontsize)
        ax.set_xlabel(xlabel, fontsize=fontsize-2)
        ax.set_ylabel(ylabel, fontsize=fontsize-2)
        ax.tick_params(axis='both', which='major', labelsize=fontsize-2)
        if bar_or == "vertical":
            cax = ax.inset_axes([1.015, 0.05, 0.03, bar_height],
                                transform=ax.transAxes)
        else:
            cax = ax.inset_axes([0.05, -0.15, bar_height, 0.05],
                                transform=ax.transAxes)
        smin = np.round(np.nanmin(data), rd)
        smax = np.round(np.nanmax(data), rd)
        ssmin = min_col
        ssmax = max_col
        ds = (ssmax-ssmin)/nticks
        ticks = np.round(np.arange(ssmin, ssmax+ds/2, ds), rd)
        ticks = list(ticks)
        cbar = self.fig.colorbar(
            im1, orientation=bar_or, ax=ax, cax=cax, fraction=0.2,
            extend="both", ticks=ticks)
        if bar_or == "vertical":
            cbar.ax.set_ylabel(clabel, size=fontsize-2)
            cbar.ax.text(0.0, -0.075, f"{smin}", verticalalignment="top",
                         horizontalalignment="left",
                         transform=cbar.ax.transAxes, fontsize=fontsize-2)
            cbar.ax.text(0.0, 1.075, f"{smax}", verticalalignment="bottom",
                         horizontalalignment="left",
                         transform=cbar.ax.transAxes, fontsize=fontsize-2)
            for lab in cbar.ax.yaxis.get_ticklabels():
                lab.set_fontsize(fontsize-2)
        else:
            cbar.ax.set_xlabel(clabel, size=fontsize-2)
            cbar.ax.text(0.0, 1.0, f"{smin}", verticalalignment="bottom",
                         horizontalalignment="right",
                         transform=cbar.ax.transAxes, fontsize=fontsize-2)
            cbar.ax.text(1.0, 1.0, f"{smax}", verticalalignment="bottom",
                         horizontalalignment="left",
                         transform=cbar.ax.transAxes, fontsize=fontsize-2)
            for lab in cbar.ax.xaxis.get_ticklabels():
                lab.set_fontsize(fontsize-2)
        if grid_flag:
            ax.grid(visible=True)
        ax.set_aspect("equal", adjustable="box")

    def plot_images(self, data, x, y, mincol=0, maxcol=0, percent=0,
                    c="rainbow", ptitle="",  xlabel="", ylabel="", clabel="",
                    grid_flag=True, bar_height=0.9):
        """
        Plot one or several 2D arrays into floating window

        Parameters
        ----------
        data : numpy array
            Data to be plotted.
            The array my be 2D or 3D. If 3D, it is supposed that several
            figures should be plotted into the same window in different frames.
            If more than one plot is to be done, the thrid dimension contains
            the different data sets. Every data set must be 2D. Several data
            sets may be concatenated into data with the following commands:

            - data = data1.reshape(nr,nc,1)
            - data = np.concatenate((data,data2.reshape(nr,nc,1)),axis=2)

            data1 and data2 are arrays with shape (nr,nc) defined on regular
            grids.

        x : numpy 1D float array
            Positions of the columns
        y : numpy 1D float array
            Positions of the rows
        wtitle : str optional, default: empty string
            Title of floating window
        mincol : float optional, default: 0
            Minimum value of color scale
        maxcol : float optional,  default: 0
            Maximum value of color scale.
            If mincol == maxcol,  and percent == 0, the limits of the color
            scale are minimum and maximum values of the data.
        percent : float optional:
            If >0, mincol and maxcol are calculated as quantiles
            (percent, 1-percent).
        c: str optional, default: "rainbow"
            color scale to be used.
        ptitle : str, optional, default: empty string
            May be a single string or a list of strings, where the length of
            the list must correspond to the length of the third dimension of
            data.
        xlabel : str, optional
            Similar as ptitle for lables of horizontal axis
        ylabel : str, optional
            Similar as ptitle for lables of vertical axis
        clabel : str, optional
            Similar as ptitle for lables of color bars

        Returns
        -------
        ax_float: List of Matplot axes

        """
# Only 2D arrays may be plotted, test if 1D data are passed to the routine
        if data.ndim < 2:
            _ = QtWidgets.QMessageBox.warning(
                None, "Warning",
                "Function plotFloating is not prepared for 1D plots",
                QtWidgets.QMessageBox.Close, QtWidgets.QMessageBox.Close)
            return False
# Create figure
        self.fig.tight_layout(w_pad=15, h_pad=2)
# If 2D array is passed create single axis
        ax_float = []
        if data.ndim == 2:
            # ax_float = self.fig.subplots(1, 1)
            self.gs = GridSpec(16, 10, self.fig)
            ax_float.append(self.fig.add_subplot(self.gs[1:15, 1:9]))
            data1 = np.copy(data)
            bar_or = "vertical"
            nticks = 10
            fontsz = 14
# If 3D array is passed create 2 or 3 axis depending on the shape of data
        else:
            ddx = x.max() - x.min()
            ddy = y.max() - y.min()
            data1 = data[:, :, 0]
# If horizontal extension is > 1.5x vertical one, plot axes vertically one
#    above the next. If not plot axis in horizontal direction
            if data.shape[2] == 3:
                facx = self.xsize / (3 * ddx)
                facy = self.ysize / (3 * ddy)
                if facx < facy:
                    self.gs = GridSpec(30, 10, self.fig)
                    ax_float.append(self.fig.add_subplot(self.gs[1:8, 1:9]))
                    ax_float.append(self.fig.add_subplot(self.gs[11:18, 1:9]))
                    ax_float.append(self.fig.add_subplot(self.gs[21:28, 1:9]))
                    bar_or = "vertical"
                    nticks = 10
                    fontsz = 10
                else:
                    self.gs = GridSpec(10, 30, self.fig)
                    ax_float.append(self.fig.add_subplot(self.gs[1:9, 1:8]))
                    ax_float.append(self.fig.add_subplot(self.gs[1:9, 11:18]))
                    ax_float.append(self.fig.add_subplot(self.gs[1:9, 21:28]))
                    bar_or = "horizontal"
                    nticks = 6
                    fontsz = 10
            else:
                facx = self.xsize / (2 * ddx)
                facy = self.ysize / (2 * ddy)
                if facx < facy:
                    self.gs = GridSpec(20, 10, self.fig)
                    ax_float.append(self.fig.add_subplot(self.gs[1:8, 1:9]))
                    ax_float.append(self.fig.add_subplot(self.gs[11:18, 1:9]))
                    bar_or = "vertical"
                    nticks = 10
                    fontsz = 12
                else:
                    self.gs = GridSpec(10, 20, self.fig)
                    ax_float.append(self.fig.add_subplot(self.gs[1:9, 1:8]))
                    ax_float.append(self.fig.add_subplot(self.gs[1:9, 11:18]))
                    bar_or = "horizontal"
                    nticks = 6
                    fontsz = 12
# Plot first into axis ax_float[0]
        if isinstance(ptitle, list):
            pt = ptitle[0]
            xl = xlabel[0]
            yl = ylabel[0]
            cl = clabel[0]
        else:
            pt = ptitle
            xl = xlabel
            yl = ylabel
            cl = clabel
        self.plot_image(ax_float[0], data1, x, y, mincol, maxcol, percent, c,
                        pt, xl, yl, cl, grid_flag, fontsz, bar_or, nticks,
                        bar_height)
# If more than one map has to be plotted, do this now
        if data.ndim > 2:
            if data.shape[2] > 1:
                self.plot_image(ax_float[1], data[:, :, 1], x, y, mincol,
                                maxcol, percent, c, ptitle[1], xlabel[1],
                                ylabel[1], clabel[1], grid_flag, fontsz,
                                bar_or, nticks, bar_height)
# If a third plot is to be done, do this now
                if data.shape[2] > 2:
                    self.plot_image(ax_float[2], data[:, :, 2], x, y, mincol,
                                    maxcol, percent, c, ptitle[1], xlabel[1],
                                    ylabel[1], clabel[1], grid_flag, fontsz,
                                    bar_or, nticks, bar_height)
        if len(ax_float) == 1:
            return ax_float[0]
        return ax_float



# -*- coding: utf-8 -*-
# Author:
#     Hermann Zeyen <hermann.zeyen@gmail.com>
#     Université Paris-Saclay, France
#
# License: BSD 3 clause

# from .io import get_files, read_geography_file, Data
# from .dialog import dialog
# from .geometrics import Geometrics

# print(f"Invoking __init__.py for {__name__}")

from .io import get_files, read_lineaments, read_synthetic_model, read_geography_file, get_mag_field, read_geometrics, write_geometrics
from .communication import get_geometry, get_time_correction, get_justify_indices, clip_parameters, get_spector1D, get_spector2D
from .dialog import Dialog
from .earth import Earth_mag
__all__ = ["get_files", "read_lineaments", "read_synthetic_model", "read_geography_file", "get_mag_field", "read_geometrics", "write_geometrics", "get_geometry", "get_time_correction", "get_justify_indices", "clip_parameters", "get_spector1D", "get_spector2D", "Dialog", "Earth_mag"]



# -*- coding: utf-8 -*-
"""
Last modified: Nov 22, 2024

@author: Hermann Zeyen
        University Paris-Saclay

Contains two classes:
    Prism: Defines position and properties of one prism
        Contains the following methods:
        - __init__
        - mag_prism (calculates magnetic effect at one point)
        - grav_prism (calculates gravity effect at one point)
        - change_props (modifies one or several properties)
        - change_coor (modifies one or several prism coordinates)

    Prism_calc: Allows defining a model composed of a series of prisms and
    calculation of effects at a vector of points.

    Contains the following methods:
        - __init__
        - add_prism: adds an entrance of class Prism to a dictionary
        - remove_prism: removes an entrance of classe Prism from
          dictionary
        - mag_forward: Solves magnetic forward problem for all prisms
          at all points
        - mag_deriv: Calculates derivatives of magnetic effect with
          respect to magnetic properties (susceptibility and/or
          remanence)
        - grav_forward: Solves gravity forward problem for all prisms
          at all points
        - grav_deriv: Calculates derivatives of magnetic effect with
          respect to density
        - create_Frechet: Assembles the derivative vectors of each
          prism into a Freceht matrix
        - get_max_prisms: Find all prisms located below the strongest
          absolute maximum of magnetic and gravity fields.
          Usually used during inversion procedure where
          the data correspond to actual misfit.
        - split: split a prism into up to 8 prisms having half size in
          all directions as long as prisms stay larger than
          a given minimum size.

"""

from . import mag_grav_utilities as utils
from ..in_out.earth import Earth_mag as Earth
import numpy as np


class Prism(Earth):
    """
    Class contains the coordinates and properties of a vertical prism with
    faces parallel to the coordinate system

    Uses collections of functions from mag_grav_utilities.py

    Input
    -----

    x: numpy float array with 2 values
        x[0]: coordinate of the western edge of the prism [m]
        x[1]: coordinate of the eastern edge of the prism [m]

    y: numpy float array with 2 values
        y[0]: coordinate of the southern edge of the prism [m]
        y[1]: coordinate of the northern edge of the prism [m]

    z: numpy float array with 2 values
        z[0]: coordinate of the upper edge of the prism [m]
        z[1]: coordinate of the lower edge of the prism [m]
        Positive downward

    sus: susceptibility [SI]

    rem: Remanent magnetization [A/m]

    inc: Inclination fo the remanent magnetization [degrees]

    dec: Declination fo the remanent magnetization [degrees]

    dens: Density of the prism [kg/m3] (not yet used, TODO)

    earth: class Earth_mag object
            contains the properties of the Earth's magnetic field

    Methods
    -------

    - __init__
    - mag_prism (calculates magnetic effect at one point)
    - grav_prism (calculates gravity effect at one point)
    - change_props (modifies one or several properties)
    - change_coor (modifies one or several prism coordinates)

    """

    def __init__(self, x, y, z, sus, rem, inc, dec, dens, earth):
        """
        x: numpy float 1D array of size 2
                coordinates of W and E faces
        y: numpy float 1D array of size 2
                coordinates of S and N faces
        z: numpy float 1D array of size 2
                coordinates of upper and lower faces (positive downward)
        sus: float
                susceptibility [SI]
        rem: float
                remanent magnetization [A/m]
        inc: float
                inclination of remanent magnetization [degrees]
        dec: float
                declination of remanent magnetization [degrees]
        dens: float
                density [kg/m3]
        earth: object of type Earth_mag found in file mag_grav_utilities.py
        """
# Susceptibility is transformed to cgs units which is used in function
# mag_prism
        self.G = 6.67e-6
        self.sus = sus/(4*np.pi)
        self.rem = rem
        self.inc = np.radians(inc)
        self.dec = np.radians(dec)
        self.rho = dens
        self.x = np.copy(x)
        self.y = np.copy(y)
        self.z = np.copy(z)
        self.earth = earth
        self.comps()
        self.iga = np.ones(2, dtype=int)
        self.iga[0] = -1
        self.igb = np.ones(2, dtype=int)
        self.igb[0] = -1
        self.igh = np.ones(2, dtype=int)
        self.igh[0] = -1
        self.der_flag_mag = False
        self.der_flag_grav = False
        self.x_neigh = []
        self.y_neigh = []
        self.z_neigh = []

    def comps(self):
        """
        Calculate magnetization components
        """
        self.tx, self.ty, self.tz = utils.magnetization_components(
            self.sus, self.rem, self.dec, self.inc, self.earth)

    def mag_prism(self, xp, yp, zp):
        """
        Calculates the exact 3D magnetic effect of one vertical prism with
        its sides parallel to the coordinate system at one measurement point.
        Needs susceptibilities in cgs system (SI/4pi))

        Parameters
        ----------
        xp, yp, zp : 1D Numpy float arrays
            Coordinates of all calculation points

        Returns
        -------
        delx, dely, delz : float
            N-S, E-W and Z component of the effect of the body

        """
# a, b and h are the distances between prism edges and calculation point
        a = np.float64(self.y-yp)
        a[abs(a) < 1.e-10] = np.float64(0.)
        isa = np.sign(a)
        a2 = a**2

        b = np.float64(self.x-xp)
        b[abs(b) < 1.e-10] = np.float64(0.)
        isb = np.sign(b)
        b2 = b**2

        ija = np.mean(isa)
        ijb = np.mean(isb)

# Make sure inside h, the coordinates are stored smaller value first
# If body is located above the calculation point, absolute values of
# z coordiantes are stored, but the sign of the anomaly will be reversed
# Z prism is positive downwards, zp (z point is positive upward)
#
# TODO:
# What happens if calculation point is located within the body or at least
# between the two limits (beside the body)?
        # ht = np.float64(abs(self.z[0]-zp))
        # hb = np.float64(abs(self.z[1]-zp))
        ht = np.float64(abs(self.z[0]+zp))
        hb = np.float64(abs(self.z[1]+zp))
        if hb < ht:
            sgh = -1.
            h = np.array([hb, ht])
        elif hb > ht:
            sgh = 1.
            h = np.array([ht, hb])
# If body thickness is zero, no need to calculate
        else:
            return 0., 0., 0.
        if h[0] < 1.e-10:
            h[0] = 0.
        h2 = h**2

        g1 = 1.
        g2 = 1.
        g3 = 1.
        t1 = 0.
        t2 = 0.
        w1 = 0.
        w2 = 0.
# Loop over all edges
        for k in range(2):
            hk = h[k]
            for i in range(2):
                ai = a[i]
                isai = isa[i]
                ad = abs(ai)
                igna = self.iga[i]
                aa = a2[i]
                for j in range(2):
                    bj = b[j]
                    isbj = isb[j]
                    ab = ai*bj
                    bb = b2[j]
                    aabb = aa+bb
                    igab = igna*self.igb[j]
                    bd = abs(bj)
                    isgn = igab*self.igh[k]
                    ft1 = 0.
                    ft2 = 0.
                    rr = aabb+h2[k]
                    if np.isclose(rr, 0.):
                        return 0., 0., 0.
                    r = np.sqrt(rr)
                    fg1 = bj+r
                    fg2 = hk+r
                    fg3 = ai+r
#
#  Needed for interior/edge points and for negative values of ai/bj that lead
#  to an undefined log (log of zero or negative value)
#
                    if np.isclose(hk, 0.) &\
                            (isbj == 0 or isai != 0 or ijb < 1):
                        if isbj == 0:
                            if ija < 0:
                                fg3 = 1./ad
                            elif ija == 0:
                                return 0., 0., 0.
                        elif isai == 0:
                            if ijb < 0:
                                fg1 = 1./bd
                            elif ijb == 0:
                                return 0., 0., 0.
                    elif abs(ab) > 0.:
                        ft1 = bj*hk/(ai*r)
                        ft2 = ai*hk/(bj*r)
#  Negative arctan and reciprocal log terms
                    if isgn != 1:
                        g1 /= fg1
                        g2 /= fg2
                        g3 /= fg3
                        ft1 = -ft1
                        ft2 = -ft2
                    else:
                        g1 *= fg1
                        g2 *= fg2
                        g3 *= fg3
                    if abs(ab) > 0.:
                        w1, t1 = utils.tandet(ft1, w1, t1)
                        w2, t2 = utils.tandet(ft2, w2, t2)
        g1 = np.log(g1)
        g2 = np.log(g2)
        g3 = np.log(g3)
        t1 = -(np.arctan(w1)+t1)
        t2 = -(np.arctan(w2)+t2)
        t3 = -(t1+t2)
        dex = (self.tx*t1+self.ty*g2+self.tz*g1)*sgh
        dey = (self.tx*g2+self.ty*t2+self.tz*g3)*sgh
        dez = (self.tx*g1+self.ty*g3+self.tz*t3)*sgh
        return dex, dey, dez

    def grav_prism(self, xp, yp, zp):
        """
        Calculates the exact 3D gravity effect of a vertical rectangular prism.

        Parameters
        ----------
        xp, yp, zp : float
            Coordinates of the point where the effect is to be calculated

        Returns
        -------
        g : float
            Gravity effect of the body

        """
        if np.isclose(self.z[0], self.z[1]):
            return 0.
        x = np.float64(self.x-xp)
        y = np.float64(self.y-yp)
        z = np.float64(self.z-zp)
        x[abs(x) < 1.e-5] = np.float64(1.e-5)
        y[abs(y) < 1.e-5] = np.float64(1.e-5)
        z[abs(z) < 1.e-5] = np.float64(1.e-5)
        x2 = x**2
        y2 = y**2
        z2 = z**2
        log_x = np.ones(2)
        log_y = np.ones(2)
        atan_z = np.zeros(2)
        sum_pi = np.zeros(2)
        sig = [1, -1]
        for k in range(2):
            for j in range(2):
                rr = z2[k]+y2[j]
                sgn = sig[j]*sig[k]
                for i in range(2):
                    s = sgn*sig[i]
                    r2 = rr+x2[i]
                    r = np.sqrt(r2)
                    if s > 0:
                        if not np.isclose(-y[j], r):
                            log_x[i] *= y[j]+r
                        if not np.isclose(-x[i], r):
                            log_y[j] *= x[i]+r
                    else:
                        if not np.isclose(-y[j], r):
                            log_x[i] /= y[j]+r
                        if not np.isclose(-x[i], r):
                            log_y[j] /= x[i]+r
                    fac = (x[i]+y[j]+r)/z[k]*s
                    atan_z[k], sum_pi[k] = utils.tandet(fac, atan_z[k],
                                                        sum_pi[k])
        g = np.sum(x*np.log(log_x) + y*np.log(log_y) +
                   2.*z*(np.arctan(atan_z)+sum_pi))
        return g*self.rho*self.G

    # def okabe(self):
        

    def change_props(self, sus=None, rem=None, inc=None, dec=None, dens=None):
        """
        Change properties of a prism for each entry that is not None

        Parameters
        ----------
        sus : float, optional; default: None
            Susceptibility [SI system].
        rem : float, optional; default: None
            Remanent magnetization [A/m].
        inc : float, optional; default: None
            Inclination of remanent magnetization [degrees].
        dec : float, optional; default: None
            Declination of remanent magnetization [degrees].
        dens : float, optional; default: None
            Density [kg/m3].

        Returns
        -------
        None.

        """
        if sus is not None:
            self.sus = sus/(4*np.pi)
        if rem is not None:
            self.rem = rem
        if inc is not None:
            self.inc = np.radians(inc)
        if dec is not None:
            self.dec = np.radians(dec)
        if dens is not None:
            self.rho = dens
        self.tx, self.ty, self.tz = utils.magnetization_components(
            sus, rem, dec, inc, self.earth)

    def change_coor(self, xmin=None, xmax=None, ymin=None, ymax=None,
                    zmin=None, zmax=None):
        """
        Change coordinates of prism faces for all entries that are not None.

        Parameters
        ----------
        xmin : float, optional; default: None
            Coordiante of western face of prism [m].
        xmax : float, optional; default: None
            Coordiante of eastern face of prism [m].
        ymin : float, optional; default: None
            Coordiante of southern face of prism [m].
        ymax : float, optional; default: None
            Coordiante of northern face of prism [m].
        zmin : float, optional; default: None
            Coordiante of upper face of prism ([m] positive below surface).
        zmax : float, optional; default: None
            Coordiante of lower face of prism ([m] positive below surface).

        Returns
        -------
        None.

        """
        if xmin is not None:
            self.x[0] = xmin
        if xmax is not None:
            self.x[1] = xmax
        if ymin is not None:
            self.y[0] = ymin
        if ymax is not None:
            self.y[1] = ymax
        if zmin is not None:
            self.z[0] = zmin
        if zmax is not None:
            self.z[1] = zmax


class Prism_calc(Prism, Earth):
    """
    Contains the following methods:

    - __init__
    - add_prism: adds an entrance of class Prism to a dictionary
    - remove_prism: removes an entrance of classe Prism from dictionary
    - mag_forward: Solves magnetic forward problem for all prisms at all points
    - mag_deriv: Calculates derivatives of magnetic effect with respect to
      magnetic properties (susceptibility and/or remanence)
    - grav_forward: Solves gravity forward problem for all prisms at all points
    - grav_deriv: Calculates derivatives of magnetic effect with respect to
      density
    - create_Frechet: Assembles the derivative vectors of each prism into a
      Frechet matrix
    - get_max_prisms: Find all prisms located below the strongest absolute
      maximum of magnetic and gravity fields. Usually used during inversion
      procedure where the data correspond to actual misfit.
    - split: split a prism into up to 8 prisms having half size in all
      directions as long as prisms stay larger than a given minimum size.

    """

    def __init__(self, e, min_size_x=0., min_size_y=0., min_size_z=0.):
        """
        Initialize dictionary of prisms used for magnetic and gravity
        calculation

        Parameters
        ----------
        e: Obejct of class Earth_mag (found in file mag_grav_utilities.py)
        min_size_x : float
            Minimum allowed size of prisms in X direction (E-W).
        min_size_y : float
            Minimum allowed size of prisms in Y direction (N-S).
        min_size_z : float
            Minimum allowed size of prisms in Z direction.

        Returns
        -------
        None.

        """
# self.G is the universal gravity constant multiplied by 10**5 so that results
#       are in mGal
        self.G = 6.67e-6
# Define dictionaray that will contain all model prisms
        self.prisms = {}
# self.n_prisms is the number of defined prisms
        self.n_prisms = 0
# self.n_max is the key of the last defined prism. Initially, n_max = n_prism.
#       However, if a prism other than the last defined one is removed from the
#       dictionary, it's key is erased, n_prism is reduced by one, but the
#       next prism that will be added to the dictionary will have key n_max+1.
#       In this case, n_prisms and n_max will be different
        self.n_max = -1
        self.min_size_x = min_size_x
        self.min_size_y = min_size_y
        self.min_size_z = min_size_z
        self.earth = e

    def add_prism(self, xpr, ypr, zpr, sus, rem, rinc, rdec, rho):
        """
        Add a prism with its properties to the dictionary.
        The key of this new prism will be self.n_max+1

        For the explanation of input parameters see Prism.__init__
        """
        self.n_max += 1
        self.prisms[self.n_max] = Prism(xpr, ypr, zpr, sus, rem, rinc, rdec,
                                        rho, self.earth)
        self.n_prisms += 1
        if self.n_prisms == 1:
            return
        x1 = xpr.min()
        x2 = xpr.max()
        y1 = ypr.min()
        y2 = ypr.max()
        z1 = zpr.min()
        z2 = zpr.max()
        keys = list(self.prisms.keys())
        for key in keys[:-1]:
            x3 = self.prisms[key].x.min()
            x4 = self.prisms[key].x.max()
            y3 = self.prisms[key].y.min()
            y4 = self.prisms[key].y.max()
            z3 = self.prisms[key].z.min()
            z4 = self.prisms[key].z.max()
            if (x3 == x2 or x4 == x1) and y2 > y3 and y1 < y4 and z2 > z3 and\
                    z1 < z4:
                self.prisms[self.n_max].x_neigh.append(key)
            if (y3 == y2 or y4 == y1) and x2 > x3 and x1 < x4 and z2 > z3 and\
                    z1 < z4:
                self.prisms[self.n_max].y_neigh.append(key)
            if (z3 == z2 or z4 == z1) and x2 > x3 and x1 < x4 and y2 > y3 and\
                    y1 < y4:
                self.prisms[self.n_max].z_neigh.append(key)

    def remove_prism(self, key):
        """
        Remove a prism for the dictionary
        n_prisms will be reduced by one, but not n_max.
        """
        try:
            del self.prisms[key]
            self.n_prisms -= 1
            for k in self.prisms.keys():
                if key in self.prisms[k].x_neigh:
                    del self.prisms[k].x_neigh[np.where(
                        np.array(self.prisms[k].x_neigh) == key)[0][0]]
                if key in self.prisms[k].y_neigh:
                    del self.prisms[k].y_neigh[np.where(
                        np.array(self.prisms[k].y_neigh) == key)[0][0]]
                if key in self.prisms[k].z_neigh:
                    del self.prisms[k].z_neigh[np.where(
                        np.array(self.prisms[k].z_neigh) == key)[0][0]]
            return True
        except KeyError:
            print(f"\nWarning in 'remove_prism': Prism {key} does not exist "
                  + "in dictionary.")
            return False

    def mag_forward(self, xp, yp, zp):
        """
        Calculate summed effect of all prisms on all field points

        Parameters
        ----------
        prisms : dictionary with Prism class objects
            Positions and properties of all prisms
        xp : numpy 1D float array [n_points]
            X-coordiante of field points
            [nr: number of rows in NS direction, nc: number columns in EW
            direction]
        yp : numpy 1D float array [n_points]
            Y-coordiante of field points
        zp : numpy 1D float array [n_points]
            Z-coordiante of field points

        Returns
        -------
        v : numpy 2D float array [n_points,5]
            Calculated anomalies.
            v[:,0] : X-component (N-S)
            v[:,1] : Y-component (E-W)
            v[:,2] : Z-component
            v[:,3] : horizontal component
            v[:,4] : total field component
        """
        n_points = len(xp)
        self.v = np.zeros((n_points, 5))
        for i in range(n_points):
            delx = 0.
            dely = 0.
            delz = 0.
            for key, val in self.prisms.items():
                dex, dey, dez = val.mag_prism(xp[i], yp[i], zp[i])
                delx += dex
                dely += dey
                delz += dez
            delh, delt = utils.compon(delx, dely, delz, self.earth)
            self.v[i, 0] += delx
            self.v[i, 1] += dely
            self.v[i, 2] += delz
            self.v[i, 3] += delh
            self.v[i, 4] += delt
            if np.mod(i+1, 100) == 0:
                print(f"Point {i+1}: Magnetic field calculated")
        return np.copy(self.v)

    def mag_deriv(self, xp, yp, zp, sus_inv=True, rem_inv=False):
        """
        Calculate summed effect of all prisms on all field points

        Parameters
        ----------
        prisms : dictionary with Prism class objects
            Positions and properties of all prisms
        xp : numpy 1D float array [n_points]
            X-coordiante of field points
            [nr: number of rows in NS direction, nc: number columns in EW
            direction]
        yp : numpy 1D float array [n_points]
            Y-coordiante of field points
        zp : numpy 1D float array [n_points]
            Z-coordiante of field points
        sus_inv : bool, optional
            If True susceptibility derivatives are calculated; Default: True
        rem_inv : bool, optional
            If True remanence derivatives are calculated; Default: False

        """
        n_points = len(xp)
        for key, val in self.prisms.items():
            if self.prisms[key].der_flag_mag:
                continue
            val.sus_der = np.zeros(n_points)
            val.rem_der = np.zeros(n_points)
            rem = val.rem
            sus = val.sus
            if sus_inv:
                val.rem = 0.
                val.sus = 1.
                val.comps()
                for i in range(n_points):
                    dex, dey, dez = val.mag_prism(xp[i], yp[i], zp[i])
                    deh, det = utils.compon(dex, dey, dez, self.earth)
                    self.prisms[key].sus_der[i] = det
                val.rem = rem
                val.sus = sus
                self.prisms[key].comps()
                self.prisms[key].der_flag_mag = True
            if rem_inv:
                val.sus = 0.
                val.rem = 1.
                val.comps()
                for i in range(n_points):
                    dexr, deyr, dezr = val.mag_prism(xp[i], yp[i], zp[i])
                    deh, det = utils.compon(dex, dey, dez, self.earth)
                    self.prisms[key].rem_der[i] = det
                val.sus = sus
                val.rem = rem
                self.prisms[key].comps()
                self.prisms[key].der_flag_mag = True
            print(f"Prism {key+1}: Magnetic derivative calculated")
        return True

    def grav_forward(self, xp, yp, zp):
        """
        Calculate summed gravity effect of all prisms on all field points

        Parameters
        ----------
        prisms : dictionary with Prism class objects
            Positions and properties of all prisms
        xp : numpy 1D float array [n_points]
            X-coordiante of field points
            [nr: number of rows in NS direction, nc: number columns in EW
            direction]
        yp : numpy 1D float array [n_points]
            Y-coordiante of field points
        zp : numpy 1D float array [n_points]
            Z-coordiante of field points

        Returns
        -------
        g : numpy 1D float array [n_points]
            Calculated anomalies.
        """
    # G is the universal gravity constant
    # in order to pass from m/s2 to mGal, it is multiplied by 10**5
        n_points = len(xp)
        self.g = np.zeros(n_points)
        for i in range(n_points):
            for key, val in self.prisms.items():
                self.g[i] += val.grav_prism(xp[i], yp[i], zp[i])
            if np.mod(i+1, 100) == 0:
                print(f"Point {i+1}: Gravi calculated")
        return np.copy(self.g)

    def grav_deriv(self, xp, yp, zp, deriv=True):
        """
        Calculate summed gravity effect of all prisms on all field points

        Parameters
        ----------
        prisms : dictionary with Prism class objects
            Positions and properties of all prisms
        xp : numpy 1D float array [n_points]
            X-coordiante of field points
            [nr: number of rows in NS direction, nc: number columns in EW
            direction]
        yp : numpy 1D float array [n_points]
            Y-coordiante of field points
        zp : numpy 1D float array [n_points]
            Z-coordiante of field points
        """
        n_points = len(xp)
        for key, val in self.prisms.items():
            if val.der_flag_grav:
                continue
            val.rho_der = np.zeros(n_points)
            rho = val.rho
            val.rho = 1.
            for i in range(n_points):
                self.prisms[key].rho_der[i] = \
                    val.grav_prism(xp[i], yp[i], zp[i])
            val.rho = rho
            self.prisms[key].der_flag_grav = True
            print(f"Prism {key+1}: Gravi derivatives calculated")
        return True

    def get_n_param(self, sus_inv, rem_inv, rho_inv):
        """
        Calculate total number of model parameters

        Parameters
        ----------
        sus_inv : bool
            If True, susceptibilities are model parameters
        rem_inv : bool
            If True, remanences are model parameters
        rho_inv : bool
            If True, densities are model parameters

        Return
        ------
            self.n_param : int
                Total number of parameters
        """
        self.n_param = 0
        if sus_inv:
            self.n_param += self.n_prisms
        if rem_inv:
            self.n_param += self.n_prisms
        if rho_inv:
            self.n_param += self.n_prisms
# add one more parameter corresponding to the unknown data zero level
        self.n_param += 1
        return self.n_param

    def create_Frechet(self, sus_inv, rem_inv, rho_inv, xp, yp, zp):
        """
        Calculate Freceht matrix

        Parameters
        ----------
        sus_inv : bool
            If True, susceptibilities are model parameters
        rem_inv : bool
            If True, remanences are model parameters
        rho_inv : bool
            If True, densities are model parameters
        xpm : 1D numpy float array (lenght: number of magnetic data points)
            X coordinates of magnetic data points
        ypm : 1D numpy float array (lenght: number of magnetic data points)
            Y coordinates of magnetic data points
        zpm : 1D numpy float array (lenght: number of magnetic data points)
            Z coordinates of magnetic data points
        xpg : 1D numpy float array (lenght: number of gravity data points)
            X coordinates of gravity data points
        ypg : 1D numpy float array (lenght: number of gravity data points)
            Y coordinates of gravity data points
        zpg : 1D numpy float array (lenght: number of gravity data points)
            Z coordinates of gravity data points

        Return
        ------
            self.Frechet : 2D numpy float matrix with size [self.n_dat,
            self.n_param]
                Frechet matrix
        """
        self.ndat = 0
        if sus_inv or rem_inv:
            n_data = len(xp)
            self.ndat += n_data
            self.mag_deriv(xp, yp, zp)
        if rho_inv:
            n_data = len(xp)
            self.ndat += n_data
            self.grav_deriv(xp, yp, zp)
        self.n_param = self.get_n_param(sus_inv, rem_inv, rho_inv)
        self.Frechet = np.zeros((self.ndat, self.n_param))
        icol = 0
        nrow = 0
        self.params = []
        if sus_inv:
            nrow1 = nrow + n_data
            for i, key in enumerate(self.prisms.keys()):
                self.Frechet[nrow:nrow1, i+icol] = self.prisms[key].sus_der
                self.params.append(self.prisms[key].sus)
            if not rem_inv:
                nrow += n_data
                nrow1 = nrow + n_data
            icol += self.n_prisms
        if rem_inv:
            for i, key in enumerate(self.prisms.keys()):
                self.Frechet[nrow:nrow1, i+icol] = self.prisms[key].rem_der
                self.params.append(self.prisms[key].rem)
            nrow += n_data
            icol += self.n_prisms
        if rho_inv:
            nrow1 = nrow + n_data
            for i, key in enumerate(self.prisms.keys()):
                self.Frechet[nrow:nrow1, i+icol] = self.prisms[key].rho_der
                self.params.append(self.prisms[key].rho)
        self.params.append(0.)
        self.Frechet[:, -1] = 1.
        self.params = np.array(self.params)
        return np.copy(self.Frechet)

    def create_smooth(self, sus_inv, rem_inv, rho_inv, sigma_sus, sigma_rem,
                      sigma_rho, depth_ref):
        """
        Create smoothing matrix. It has size [n_params,n_params].
        On the diagonal is the number of neighbouring blocks (2 to 4). at the
        position [i,j] value is -1. i,j are numbers of neighbouring blocks (not
        block keys but position of the block in the dictionary)
        In the actual version, no smoothing in Z is done

        Parameters
        ----------
        sus_inv : bool
            If True, susceptibilities are model parameters
        rem_inv : bool
            If True, remanences are model parameters
        rho_inv : bool
            If True, densities are model parameters
        sigma_sus : float
            Variability of susceptibility [SI units]
        sigma_rem : float
            Variability of remances [A/m]
        sigma_rho : float
            Variability of densities [kg/m3]
        depth_ref : float
            Reference depth to increase smoothing with depth

        Returns
        -------
        S : 2D numpy float array
            Smoothing matrix mith size [n_params,n_params].

        """
        S = np.zeros((self.n_param, self.n_param))
        SS = np.zeros((self.n_prisms, self.n_prisms))
        z_fac = np.zeros(self.n_prisms)
        keys = np.array(list(self.prisms.keys()))
# Check whether neighbouring block in X direction exists
        for i, key in enumerate(keys):
            z_fac[i] = (self.prisms[key].z[0] + self.prisms[key].z[1])\
                / (2*depth_ref)
# Determine numbers of neighbouring blocks in dictionary
            if len(self.prisms[key].x_neigh) > 0:
                for kk in self.prisms[key].x_neigh:
                    k = np.where(keys == kk)[0]
# Add one to the diagonal positions of each block and set the position [i,k]
#     and the symmetric one [k,i] to -1
                    SS[i, i] += 1.
                    SS[k, k] += 1.
                    SS[k, i] = -1.
                    SS[i, k] = -1.
# Do the same for neighbouring blocks in Y direction
            if len(self.prisms[key].y_neigh) > 0:
                for kk in self.prisms[key].y_neigh:
                    k = np.where(keys == kk)[0]
                    SS[i, i] += 1.
                    SS[k, k] += 1.
                    SS[k, i] = -1.
                    SS[i, k] = -1.

# If inversion is done for at least two different parameter types, set the
#    smmothing parameters for the second parameter type
        z_fac[:] = 1.
        n0 = 0
        n1 = self.n_prisms
        if sus_inv:
            S[n0:n1, n0:n1] = SS/sigma_sus**2
            for i in range(len(keys)):
                S[i, :] *= z_fac[i]
                S[:, i] *= z_fac[i]
            n0 = n1
            n1 += self.n_prisms
        if rem_inv:
            S[n0:n1, n0:n1] = SS/sigma_rem**2
            for i in range(len(keys)):
                S[i+n0, :] *= z_fac[i]
                S[:, i+n0] *= z_fac[i]
            n0 = n1
            n1 += self.n_prisms
        if rho_inv:
            S[n0:n1, n0:n1] = SS/sigma_rem**2
            for i in range(len(keys)):
                S[i+n0, :] *= z_fac[i]
                S[:, i+n0] *= z_fac[i]
        return S

    def get_max_prisms(self, data, deriv, max_lim=0.1, width=10):
        """
        Find maxima in the matrix data defined as points that are larger than
        all other points in an area if width points around. For maxima with
        values larger than
        maxlim*(maximum_of_maxima - minimum_of_maxima)+minimum_of_maxima
        the prism is detected that has the strongest influence on the value of
        of the corresponding point (maximum derivative). A minimum of 3 prisms
        is searched for.
        If splitting the prism results in a prism smaller than the limit
        defined during initialization of the Class, do not integrate the
        corresponding prism in the outputted list

        Parameters
        ----------
        data : numpy 2D array of shape (ny,nx); float
            data set from which to find the extreme value.
        deriv : numpy 2D array (ndata x n_parameters), ndata = nx*ny
            Frechet matrix linking data to parameters (if several classes of
            data and/or parameters are used, pass only the part corresponding
            to one data/parameter class, i.e. pass a partial Frechet matrix)
        max_lim : float, optional Default is 0.1
            Prisms having the strongest influence on data values larger than
            the formula given above will be split.
        width : int, optional
            A maximum is recognized if the value at a point (i,j) is larger
            than or equal to all other values within an area of
            (i-width:i+width, j-width:j+width)


        Returns
        -------
        keys_max : list; int
            key values of all prisms lying beneath the maximum.

        """
        keys = list(self.prisms.keys())
        keys_max = []
        if len(data.shape) == 1:
            nx = len(data)
            ny = 1
        else:
            ny, nx = data.shape
        _, maxima = utils.get_extremes(abs(data), width)
        maxima = np.array(maxima)
        max_val = np.nanmin(data)
        min_val = np.nanmax(data)
        max_values = []
        max_keys = []
        for pos in maxima:
            npos = len(pos)
            if npos == 1:
                i = pos[0]
            else:
                i = pos[0]*nx + pos[1]
            p = np.argmax(abs(deriv[i, :]))
            key = keys[p]
            xpmn = self.prisms[key].x[0]
            xpmx = self.prisms[key].x[1]
            ypmn = self.prisms[key].y[0]
            ypmx = self.prisms[key].y[1]
            zpmn = self.prisms[key].z[0]
            zpmx = self.prisms[key].z[1]
            dx = (xpmx-xpmn)*0.5
            dy = (ypmx-ypmn)*0.5
            dz = (zpmx-zpmn)*0.5
# Check whether prims corresponding to this maximum can be split further
            if dx >= self.min_size_x or dy >= self.min_size_y or\
                    dz >= self.min_size_z:
                if npos == 1:
                    max_values.append(data[i])
                else:
                    max_values.append(data[pos[0], pos[1]])
                max_val = max(max_val, max_values[-1])
                min_val = min(min_val, max_values[-1])
                max_keys.append(key)
        max_values = np.array(max_values)
        max_keys = np.array(max_keys)
# Calculate the limiting value of maxima to be used for splitting of prisms
        max_test = min_val + (max_val - min_val)*max_lim
        index = np.argsort(max_values)
        max_keys = max_keys[index]
        max_values = max_values[index]
# If one local maximum is much larger than all the others, search for the
# value of the third largest local maximum and use this value as limit
#        npris = 0
        # for pos in maxima[index[::-1]]:
        #     if data[pos[0], pos[1]] < max_test and npris > 2:
        #         continue
        #     i = pos[0]*nx + pos[1]
        #     p = np.argmax(abs(deriv[i, :]))
        #     key = keys[p]
        #     xpmn = self.prisms[key].x[0]
        #     xpmx = self.prisms[key].x[1]
        #     ypmn = self.prisms[key].y[0]
        #     ypmx = self.prisms[key].y[1]
        #     zpmn = self.prisms[key].z[0]
        #     zpmx = self.prisms[key].z[1]
        #     dx = (xpmx-xpmn)*0.5
        #     dy = (ypmx-ypmn)*0.5
        #     dz = (zpmx-zpmn)*0.5
        #     if dx >= self.min_size_x or dy >= self.min_size_y or\
        #             dz >= self.min_size_z:
        #         keys_max.append(key)
        #         npris += 1
        for i, m in enumerate(max_values):
            if m >= max_test:
                keys_max.append(max_keys[i])
#            npris += 1
        return list(np.unique(np.array(keys_max)))

    def split(self, key):
        """
        Replace a prism by up to eight prisms, half the size in all directions.
        In each direction, the splitted size must not be smaller than the
        minimum size defined during initialization of the Class.

        Parameters
        ----------
        key : int
            key of the prism to be split in the dictionary.


        """
        sus_act = self.prisms[key].sus*4.*np.pi
        rem_act = self.prisms[key].rem
        rho_act = self.prisms[key].rho
        inc_act = self.prisms[key].inc
        dec_act = self.prisms[key].dec
        xpmn = self.prisms[key].x[0]
        xpmx = self.prisms[key].x[1]
        dx_act = xpmx-xpmn
        ypmn = self.prisms[key].y[0]
        ypmx = self.prisms[key].y[1]
        dy_act = ypmx-ypmn
        zpmn = self.prisms[key].z[0]
        zpmx = self.prisms[key].z[1]
        dz_act = zpmx-zpmn
        dx_p = dx_act/2.
        if dx_p < self.min_size_x:
            xp0 = [xpmn]
            xp1 = [xpmx]
        else:
            xp0 = [xpmn, xpmn+dx_p]
            xp1 = [xpmn+dx_p, xpmx]
        dy_p = dy_act/2.
        if dy_p < self.min_size_y:
            yp0 = [ypmn]
            yp1 = [ypmx]
        else:
            yp0 = [ypmn, ypmn+dy_p]
            yp1 = [ypmn+dy_p, ypmx]
        dz_act = zpmx-zpmn
        dz_p = dz_act/2.
        if dz_p < self.min_size_z:
            zp0 = [zpmn]
            zp1 = [zpmx]
        else:
            zp0 = [zpmn, zpmn+dz_p]
            zp1 = [zpmn+dz_p, zpmx]
        self.remove_prism(key)
        key_add = []
        for ix in range(len(xp0)):
            xpr = np.array([xp0[ix], xp1[ix]])
            for iy in range(len(yp0)):
                ypr = np.array([yp0[iy], yp1[iy]])
                for iz in range(len(zp0)):
                    zpr = np.array([zp0[iz], zp1[iz]])
                    self.add_prism(xpr, ypr, zpr, sus_act, rem_act, inc_act,
                                   dec_act, rho_act)
                    key_add.append(self.n_max)
        return key_add



# -*- coding: utf-8 -*-
"""
Created on Sun Sep  8 21:27:45 2024

@author: Hermann
"""

import numpy as np

class Earth_mag():
    """
    
    Class stores the components of the Earth's magnetic field and the direction
    (co)sines
    
    """

    def __init__(self, intensity, inclination, declination):
        self.f = intensity
        self.inc = inclination
        self.dec = declination
        self.earth_components()

    def earth_components(self):
        self.cde = np.cos(np.radians(self.dec))
        self.sde = np.sin(np.radians(self.dec))
        self.cie = np.cos(np.radians(self.inc))
        self.sie = np.sin(np.radians(self.inc))
        self.cdci = self.cde*self.cie
        self.sdci = self.sde*self.cie
        self.eh = self.f*self.cie
        self.ex = self.f*self.cdci
        self.ey = self.f*self.sdci
        self.ez = self.f*self.sie
        return True

